'''
Functions that given a TargetCollection and any number of kwargs will 
return a list of `target_id`s from the given TargetCollection.

These functions are used to subset TargetCollection so that we can get 
subset metrics e.g. Accuracy score for samples within the TargetCollection 
that have only one sentiment within the sentence that sample came from
'''
from collections import defaultdict
from typing import Dict, Any, Set, List

from bella.data_types import TargetCollection, Target

def targets_to_samples(dataset: TargetCollection, targets: Set[str], 
                       lower: bool = True) -> List[Target]:
    '''
    Given a dataset and a set of target words, it will return a subset of the 
    dataset where all samples in the subset have target words that are in the 
    targets set.

    :param dataset: TargetCollection containing samples
    :param targets: A set of target words used to subset the dataset
    :param lower: Whether to lower case the target words. If this is True 
                  it is up to you to ensure all the words in the `targets` set 
                  have been lower cased.
    :returns: A subset of the dataset where all targets in the subset are 
              within the `targets` set.
    '''
    samples = []
    for data in dataset.data():
        target = data['target']
        if lower:
            target = target.lower()
        if target in targets:
            samples.append(data)
    return samples

def target_sentiments(dataset: TargetCollection, 
                      lower: bool = True) -> Dict[str, Set[Any]]:
    '''
    Given a dataset will return a dictionary of targets and the sentiment 
    that has been associated to those targets.

    E.g. within the dataset that `target` `camera` may have only been seen 
    with a positive and a negative label but not neutral therefore in the 
    returned dictionary it would be {`camera`: [`positive`, `negative`]}

    :param dataset: TargetCollection containing samples
    :param lower: Whether to lower case the target words.
    :returns: A dictionary where the keys are target words and the values 
              are the sentiment values that have been associated to those 
              targets.
    '''
    targets_sentiments = defaultdict(set)
    for data in dataset.data():
        target = data['target']
        if lower:
            target = target.lower()
        targets_sentiments[target].add(data['sentiment'])
    return targets_sentiments

def same_one_sentiment(test_dataset: TargetCollection, 
                       train_dataset: TargetCollection, 
                       lower: bool = True) -> List[str]:
    '''
    Given a test and train dataset will return all of the test dataset sample 
    ids that contain targets that have only occured once in the train and test 
    sets with the same sentiment.

    :param test_dataset: Test TargetCollection
    :param train_dataset: Train TargetCollection
    :param lower: Whether to lower case the target words
    :returns: A list of sample ids from the test dataset.
    '''
    train_target_sentiments = target_sentiments(train_dataset, lower)
    test_target_sentiments = target_sentiments(test_dataset, lower)

    same_one_sentiments = set()
    for data in test_dataset.data():
        target = data['target']
        if lower:
            target = target.lower()
        if (target in train_target_sentiments and 
            target in test_target_sentiments):
            train_sentiments = train_target_sentiments[target]
            test_sentiments = test_target_sentiments[target]
            if (len(train_sentiments) == 1 and 
                len(test_sentiments) == 1):
                if train_sentiments == test_sentiments:
                    same_one_sentiments.add(target)

    same_one_samples = targets_to_samples(test_dataset, same_one_sentiments, 
                                          lower)
    same_one_ids = [sample['target_id'] for sample in same_one_samples]
    return same_one_ids

def same_multi_sentiment(test_dataset: TargetCollection,
                         train_dataset: TargetCollection,
                         lower: bool = True) -> List[str]:
    '''
    Given a test and train dataset will return all of the test dataset sample 
    ids that contain targets that have occured more than once in the train and 
    test sets with the same sentiment labels.

    :param test_dataset: Test TargetCollection
    :param train_dataset: Train TargetCollection
    :param lower: Whether to lower case the target words
    :returns: A list of sample ids from the test dataset.
    '''
    train_target_sentiments = target_sentiments(train_dataset, lower)
    test_target_sentiments = target_sentiments(test_dataset, lower)

    same_multi_sentiments = set()
    for data in test_dataset.data():
        target = data['target']
        if lower:
            target = target.lower()
        if (target in train_target_sentiments and 
            target in test_target_sentiments):
            train_sentiments = train_target_sentiments[target]
            test_sentiments = test_target_sentiments[target]
            if (len(train_sentiments) > 1 and 
                len(test_sentiments) > 1):
                if train_sentiments == test_sentiments:
                    same_multi_sentiments.add(target)
    same_multi_samples = targets_to_samples(test_dataset, same_multi_sentiments, 
                                            lower)
    same_multi_ids = [sample['target_id'] for sample in same_multi_samples]
    return same_multi_ids

def similar_sentiment(test_dataset: TargetCollection,
                      train_dataset: TargetCollection,
                      lower: bool = True) -> List[str]:
    '''
    Given a test and train dataset will return all of the test dataset sample 
    ids that contain targets that have occured more than once in the train or 
    test sets with at least some overlap between the test sentiment and train 
    but not identical. E.g. the target `camera` could occur with `positive` and
    `negative` sentiment in the test set and only `negative` in the train set.

    :param test_dataset: Test TargetCollection
    :param train_dataset: Train TargetCollection
    :param lower: Whether to lower case the target words
    :returns: A list of sample ids from the test dataset.
    '''
    train_target_sentiments = target_sentiments(train_dataset, lower)
    test_target_sentiments = target_sentiments(test_dataset, lower)

    similar_sentiments = set()
    for data in test_dataset.data():
        target = data['target']
        if lower:
            target = target.lower()
        if (target in train_target_sentiments and 
            target in test_target_sentiments):
            train_sentiments = train_target_sentiments[target]
            test_sentiments = test_target_sentiments[target]
            if (len(train_sentiments) > 1 or 
                len(test_sentiments) > 1):
                if train_sentiments == test_sentiments:
                    continue
                if test_sentiments.intersection(train_sentiments):
                    similar_sentiments.add(target)
    similar_samples = targets_to_samples(test_dataset, similar_sentiments, 
                                         lower)
    similar_ids = [sample['target_id'] for sample in similar_samples]
    return similar_ids

def different_sentiment(test_dataset: TargetCollection,
                        train_dataset: TargetCollection, 
                        lower: bool = True) -> List[str]:
    '''
    Given a test and train dataset will return all of the test dataset sample 
    ids that contain targets that have different sentiment labels with no 
    overlap in the test compared to the train set.

    :param test_dataset: Test TargetCollection
    :param train_dataset: Train TargetCollection
    :param lower: Whether to lower case the target words
    :returns: A list of sample ids from the test dataset.
    '''
    train_target_sentiments = target_sentiments(train_dataset, lower)
    test_target_sentiments = target_sentiments(test_dataset, lower)

    different_sentiments = set()
    for data in test_dataset.data():
        target = data['target']
        if lower:
            target = target.lower()
        if (target in train_target_sentiments and 
            target in test_target_sentiments):
            train_sentiments = train_target_sentiments[target]
            test_sentiments = test_target_sentiments[target]
            if not test_sentiments.intersection(train_sentiments):
                different_sentiments.add(target)
    different_samples = targets_to_samples(test_dataset, different_sentiments, 
                                           lower)
    different_ids = [sample['target_id'] for sample in different_samples]
    return different_ids

def unknown_targets(test_dataset: TargetCollection,
                    train_dataset: TargetCollection, 
                    lower: bool = True) -> List[str]:
    '''
    Given a test and train dataset will return all of the test dataset sample 
    ids that contain targets that did not exist in the training data.

    :param test_dataset: Test TargetCollection
    :param train_dataset: Train TargetCollection
    :param lower: Whether to lower case the target words
    :returns: A list of sample ids from the test dataset.
    '''
    train_target_sentiments = target_sentiments(train_dataset, lower)
    test_target_sentiments = target_sentiments(test_dataset, lower)

    unknowns = set()
    for data in test_dataset.data():
        target = data['target']
        if lower:
            target = target.lower()
        if (target in train_target_sentiments and 
            target in test_target_sentiments):
            continue
        else:
            unknowns.add(target)

    unknown_samples = targets_to_samples(test_dataset, unknowns, lower)
    unknown_ids = [sample['target_id'] for sample in unknown_samples]
    return unknown_ids


import numpy as np
from sklearn.base import TransformerMixin
from sklearn.base import BaseEstimator

from bella.word_vectors import WordVectors

class ContextWordVectors(BaseEstimator, TransformerMixin):

    def __init__(self, vectors=None, zero_token='$$$ZERO_TOKEN$$$'):
        self.vectors = vectors
        self.zero_token = zero_token

    def fit(self, context_tokens, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, context_tokens, y=None):
        '''see self.transform'''

        return self.transform(context_tokens)

    def transform(self, contexts_tokens):
        '''
        Given a list of contexts (either right, left or target) which are made
        up of lists of tokens return the tokens as a word vector matrix.

        The word vector matrix is a word vector for each token but instead of
        storing in a list it is stored in a numpy.ndarray of shape:
        (length of word vector, number of tokens).

        Example of the input
        [[['context', 'one'], ['context', 'two']], [['another context']]]

        :param contexts_tokens: A list of data of which each data contains a list \
        of contexts which contains a list of tokens.
        :type context_tokens: list
        :returns: The same list but with word vectors as numpy.ndarray instead \
        of tokens which are Strings
        :rtype: list
        '''
        context_word_vectors = []
        for context_tokens in contexts_tokens:
            all_contexts = []
            for context in context_tokens:
                context_word_vector = []
                for token in context:
                    token_vector = []
                    for word_vector in self.vectors:
                        token_vector.append(word_vector.lookup_vector(token))
                    context_word_vector.append(np.hstack(token_vector))
                # Padding
                if len(context_word_vector) == 0:
                    token_vector = []
                    for word_vector in self.vectors:
                        token_vector.append(word_vector.lookup_vector(self.zero_token))
                    context_word_vector.append(np.hstack(token_vector))

                context_matrix = self.list_to_matrix(context_word_vector)
                all_contexts.append(context_matrix)
            context_word_vectors.append(all_contexts)
        return context_word_vectors

    @staticmethod
    def list_to_matrix(word_vector_list):
        '''
        Converts a list of numpy.ndarrays (vectors) into a numpy.ndarray (matrix).

        :param word_vector_list: list of numpy.ndarray
        :type word_vector_list: list
        :returns: a matrix of the numpy.ndarray
        :rtype: numpy.ndarray
        '''

        num_rows = len(word_vector_list)
        matrix = np.asarray(word_vector_list)
        if matrix.shape[0] != num_rows:
            raise ValueError('The matrix row should equal the number of tokens')
        return matrix



'''
Contains functions that get different contexts from a text based on where the
target of the sentiment is within the text.


Functions:

1. right_context
2. left_context
3. target_context
4. full_context

Each function above relies on the private `_context` function.
'''

def context(target_dict, specific_context, inc_target=False):
    '''
    Returns a list of Strings based on the location of the target word in the
    text within the target dict (NOTE the target word can occur more than once
    hence why a list is returned as the context is returned for each occurence).
    specific_context can be one of the following:

    1. left - left of the target occurence.
    2. right - right of the target occurence.
    3. target - target word/words of each target occurence.
    4. full - whole text repeated for each occurence.

    The list will be length 1 if the target word only occurs once in the text.

    :param target_dict: Dictionary that contains text and the spans of the \
    target word in the text.
    :param specific_context: String specifying either the context e.g. left.
    :param inc_target: Whether to include the target word in the context text. \
    (Only applies for left and right context.)
    :type target_dict: dict
    :type specific_context: String
    :type inc_target: Boolean Default False
    :returns: A list of context strings
    :rtype: list
    '''

    acceptable_contexts = {'left', 'right', 'target', 'full'}
    if specific_context not in acceptable_contexts:
        raise ValueError('context parameter can only be one of the following {}'\
                         ' not {}'.format(acceptable_contexts, specific_context))
    text = target_dict['text']
    spans = target_dict['spans']
    contexts = []
    for span in spans:
        start_char = span[0]
        end_char = span[1]
        if specific_context == 'left':
            if inc_target:
                contexts.append(text[:end_char])
            else:
                contexts.append(text[:start_char])
        elif specific_context == 'right':
            if inc_target:
                contexts.append(text[start_char:])
            else:
                contexts.append(text[end_char:])
        elif specific_context == 'target':
            contexts.append(text[start_char:end_char])
        elif specific_context == 'full':
            contexts.append(text)
        else:
            raise ValueError('context parameter should only be `right` or '\
                             '`left` not {} there must be a logic error'\
                             .format(specific_context))
    return contexts



'''
Module contains all of the main base classes for the machine learning models
these are grouped into 3 categories; 1. Mixin, 2. Abstract, and 3. Concrete.


Mixin classes - This is a function based class that contains functions that
do not rely on the type of model and are useful for all:

1. :py:class:`bella.models.base.ModelMixin`

Abstract classes - This is used to enforce all the functions that all
the machine learning models must have. This is also the class that inherits
the Mixin class:

1. :py:class:`bella.models.base.BaseModel`

Concrete classes - These are more concete classes that still contain some
abstract methods. However they are the classes to inherit from to create a
machine learning model base on a certain framework e.g. SKlearn or Keras:

1. :py:class:`bella.models.base.SKLearnModel`
2. :py:class:`bella.models.base.KerasModel`
'''

from abc import ABC, abstractmethod
from collections import defaultdict
import copy
import os
from pathlib import Path
import pickle
import random as rn
import tempfile
from typing import Any, List, Dict, Union, Tuple, Callable
from multiprocessing.pool import Pool

import keras
from keras.callbacks import EarlyStopping, ModelCheckpoint
from keras import preprocessing
import numpy as np
import pandas as pd
import sklearn
from sklearn.externals import joblib
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn.model_selection import GridSearchCV, StratifiedKFold
import tensorflow as tf

import bella
from bella.data_types import TargetCollection, Target


class ModelMixin():
    '''
    Mixin class for all of the machine learning models. Contain functions
    only so they are as generic as possible.

    Functions:

    1. train_val_split -- Splits the training dataset into a train and
       validation set in a stratified split.
    '''

    @staticmethod
    def _convert_to_targets(data: List[Dict[str, Any]]
                            ) -> List['bella.data_types.Target']:
        '''
        Converts a list of dictionaries into a list of
        :py:class:`bella.data_types.Target`.
        '''

        all_targets = []
        for target in data:
            all_targets.append(Target(**target))
        return all_targets

    @staticmethod
    def train_val_split(train: 'TargetCollection',
                        split_size: float = 0.2, seed: Union[None, int] = 42
                        ) -> Tuple[Tuple[np.ndarray, np.ndarray],
                                   Tuple[np.ndarray, np.ndarray]]:
        '''
        Splits the training dataset into a train and validation set in a
        stratified split.

        :param train: The training dataset that needs to be split into
        :param split_size: Fraction of the dataset to assign to the
                           validation set.
        :param seed: Seed value to give to the stratified splitter. If
                     None then it uses the radnom state of numpy.
        :return: Two tuples of length two where each tuple is the train
                 and validation splits respectively, and each tuple contains
                 the data (X) and class labels (y) respectively. Returns
                 ((X_train, y_train), (X_val, y_val))
        '''
        splitter = StratifiedShuffleSplit(n_splits=1, test_size=split_size,
                                          random_state=seed)
        data = np.asarray(train.data_dict())
        sentiment = np.asarray(train.sentiment_data())
        for train_indexs, test_indexs in splitter.split(data, sentiment):
            train_data = data[train_indexs]
            test_data = data[test_indexs]

        train = TargetCollection(ModelMixin._convert_to_targets(train_data))
        val = TargetCollection(ModelMixin._convert_to_targets(test_data))

        X_train = np.array(train.data_dict())
        y_train = np.array(train.sentiment_data())
        X_val = np.array(val.data_dict())
        y_val = np.array(val.sentiment_data())
        return (X_train, y_train), (X_val, y_val)


class BaseModel(ModelMixin, ABC):
    '''
    Abstract class for all of the machine learning models.

    Attributes:

    1. model -- Machine learning model that is associated to this instance.
    2. fitted -- If the machine learning model has been fitted (default False)

    Methods:

    1. fit -- Fit the model according to the given training data.
    2. predict -- Predict class labels for samples in X.
    3. probabilities -- The probability of each class label for all samples
       in X.
    4. __repr__ -- Name of the machine learning model.

    Class Methods:

    1. name -- -- Returns the name of the model.

    Functions:

    1. save -- Saves the given machine learning model instance to a file.
    2. load -- Loads the entire machine learning model from a file.
    3. evaluate_parameter -- fit and predict given training, validation and
       test data the given model when the given parameter is changed on the
       model.
    4. evaluate_parameters -- same as evaluate_parameter however it
       evaluates over many parameter values for the same parameter.
    '''

    @abstractmethod
    def fit(self, X: np.ndarray, y: np.ndarray) -> None:
        '''
        Fit the model according to the given training data.

        :param X: Training samples matrix, shape = [n_samples, n_features]
        :param y: Training targets, shape = [n_samples]
        :return: The `model` attribute will now be trained.
        '''

    @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        '''
        Predict class labels for samples in X.

        :param X: Test samples matrix, shape = [n_samples, n_features]
        :return: Predicted class label per sample, shape = [n_samples]
        '''

    @abstractmethod
    def probabilities(self, X: np.ndarray) -> np.ndarray:
        '''
        The probability of each class label for all samples in X.

        :param X: Test samples matrix, shape = [n_samples, n_features]]
        :return: Probability of each class label for all samples, shape = \
        [n_samples, n_classes]
        '''

    @abstractmethod
    def __repr__(self) -> str:
        '''
        Name of the machine learning model.

        :return: Name of the machine learning model.
        '''

    @staticmethod
    @abstractmethod
    def save(model: 'BaseModel', save_fp: Path) -> None:
        '''
        Saves the entire machine learning model to a file.

        :param model: The machine learning model instance to be saved.
        :param save_fp: File path of the location that the model is to be \
        saved to.
        :return: Nothing.
        '''

    @staticmethod
    @abstractmethod
    def load(load_fp: Path) -> 'bella.models.base.BaseModel':
        '''
        Loads the entire machine learning model from a file.

        :param load_fp: File path of the location that the model was saved to.
        :return: self
        '''

    @staticmethod
    @abstractmethod
    def evaluate_parameter(model: 'BaseModel',
                           train: Tuple[np.ndarray, np.ndarray],
                           val: Union[None, Tuple[np.ndarray, np.ndarray]],
                           test: np.ndarray, parameter_name: str,
                           parameter: Any) -> Tuple[Any, np.ndarray]:
        '''
        Given a model will set the `parameter_name` to `parameter` fit the
        model and return the a Tuple of parameter changed and predictions of
        the model on the test data, using the train and validation data for
        fitting.

        :param model: :py:class:`bella.models.base.BaseModel` instance
        :param train: Tuple of `(X_train, y_train)`. Used to fit the model.
        :param val: Tuple of `(X_val, y_val)` or None is not required.
                    This is only required if the model requires validation
                    data like the :py:class:`bella.models.base.KerasModel`
                    models do.
        :param test: `X_test` data to predict on.
        :param parameter_name: Name of the parameter to change e.g. optimiser
        :param parameter: value to assign to the parameter e.g.
                          :py:class:`keras.optimizers.RMSprop`
        :return: A tuple of (parameter value, predictions)
        '''

    @staticmethod
    @abstractmethod
    def evaluate_parameters(model: 'bella.models.base.BaseModel',
                            train: Tuple[np.ndarray, np.ndarray],
                            val: Union[None, Tuple[np.ndarray, np.ndarray]],
                            test: np.ndarray, parameter_name: str,
                            parameters: List[Any], n_jobs: int
                            ) -> List[Tuple[Any, np.ndarray]]:
        '''
        Performs :py:func:`bella.models.base.BaseModel.evaluate_parameter` on
        one `parameter_name` but with multiple parameter values.

        This is useful if you would like to know the affect of changing the
        values of a parameter. It can also perform the task in a
        multiprocessing manner if `n_jobs` > 1.

        :param model: :py:class:`bella.models.base.BaseModel` instance
        :param train: Tuple of `(X_train, y_train)`. Used to fit the model.
        :param val: Tuple of `(X_val, y_val)` or None is not required.
                    This is only required if the model requires validation
                    data like the :py:class:`bella.models.base.KerasModel`
                    models do.
        :param test: `X_test` data to predict on.
        :param parameter_name: Name of the parameter to change e.g. optimiser
        :param parameters: A list of values to assign to the parameter e.g.
                           [:py:class:`keras.optimizers.RMSprop`]
        :param n_jobs: Number of cpus to use for multiprocessing if 1 then
                       will not multiprocess.
        :return: A list of tuples of (parameter value, predictions)
        '''

    @classmethod
    @abstractmethod
    def name(cls) -> str:
        '''
        Returns the name of the model.

        :return: Name of the model
        '''

    @property
    def model(self) -> Any:
        '''
        Machine learning model that is associated to this instance.

        :return: The machine learning model
        '''

        return self._model

    @model.setter
    def model(self, value) -> None:
        '''
        Sets the model attribute

        :param value: The value to assign to the model attribute
        '''

        self._model = value

    @property
    def fitted(self) -> bool:
        '''
        If the machine learning model has been fitted (default False)

        :return: True or False
        '''

        return self._fitted

    @fitted.setter
    def fitted(self, value: bool) -> None:
        '''
        Sets the fitted attribute

        :param value: The value to assign to the fitted attribute
        '''

        self._fitted = value


class KerasModel(BaseModel):
    '''
    Concrete class that is designed to be used as the base class for all
    machine learning models that are based on the
    `Keras library <https://keras.io>`_.

    Attributes:

    1. tokeniser -- Tokeniser model uses e.g. :py:meth:`str.split`.
    2. embeddings -- the word embeddings the model uses. e.g.
       :py:class:`bella.word_vectors.SSWE`
    3. lower -- if the model lower cases the words when pre-processing the data
    4. reproducible -- Whether to be reproducible. If None then it is quicker
       to run. Else provide a `int` that will represent the random seed value.
    5. patience -- Number of epochs with no improvement before training
       is stopped.
    6. batch_size -- Number of samples per gradient update.
    7. epcohs -- Number of times to train over the entire training set
       before stopping.
    8. optimiser -- Optimiser the model uses.
       e.g. :py:class:`keras.optimizers.SGD`
    9. optimiser_params -- Parameters for the optimiser. If None uses default
       for the optimiser being used.

    Abstract Methods:

    1. keras_model -- Keras machine Learning model that represents the class
       e.g. single forward LSTM.
    2. create_training_text -- Converts the training and validation data into
       a format that the keras model can take as input.
    3. create_training_y -- Converts the training and validation targets into a
       format that can be used by the keras model.

    Methods:

    1. fit -- Fit the model according to the given training and validation
       data.
    2. probabilities -- The probability of each class label for all samples
       in X.
    3. predict -- Predict class labels for samples in X.

    Functions:

    1. save -- Given a instance of this class will save it to a file.
    2. load -- Loads an instance of this class from a file.
    3. evaluate_parameter -- fit and predict given training, validation and
       test data the given model when the given parameter is changed on the
       model.
    4. evaluate_parameters -- same as evaluate_parameter however it
       evaluates over many parameter values for the same parameter.
    '''

    @abstractmethod
    def keras_model(self, num_classes: int) -> 'keras.models.Model':
        '''
        Keras machine Learning model that represents the class e.g.
        single forward LSTM.

        :returns: Keras machine learning model
        '''
        pass

    @abstractmethod
    def create_training_text(self, train_data: List[Dict[str, Any]],
                             validation_data: List[Dict[str, Any]]
                             ) -> Tuple[Any, Any]:
        '''
        Converts the training and validation data into a format that the keras
        model can take as input.

        :return: A tuple of length two containing the keras model training and
                 validation input respectively.
        '''

    @abstractmethod
    def create_training_y(self, train_y: np.ndarray, validation_y: np.ndarray
                          ) -> Tuple[np.ndarray, np.ndarray]:
        '''
        Converts the training and validation targets into a format that can
        be used by the keras model

        :return: A tuple of length containing two array the first for
                 training and the second for validation.
        '''

    @abstractmethod
    def _pre_process(self, data_dicts: Dict[str, Any], training: bool):
        '''
        Converts the training or validation data into a format that will be
        used by the keras model.

        This function is normally used to process the training and the
        validation to be returned together by
        :py:meth:`bella.models.base.KerasModel.create_training_text`
        '''

    def process_text(self, texts: List[str], max_length: int,
                     padding: str = 'pre', truncate: str = 'pre'
                     ) -> Tuple[int, np.ndarray]:
        '''
        Given a list of Strings, tokenised the text and lower case if set and
        then convert the tokens into a integers representing the tokens in the
        embeddings. Lastly it pads the data based on the max_length param.

        If the max_length is smaller than the sentences size it truncates the
        sentence. If max_length = -1 then the max_length is that of the longest
        sentence in the texts.

        :params texts: List of texts
        :params max_length: How many tokens a sentence can contain. If it is
                            -1 then it uses the sentence with the most tokens
                            as the max_length parameter.
        :params padding: Which side of the sentence to pad: `pre` beginning,
                         `post` end.
        :params truncate: Which side of the sentence to truncate: `pre`
                          beginning `post` end.
        :returns: A tuple of length 2 containg: 1. The max_length parameter,
                  2. A matrix of shape [n_samples, pad_size] where each integer
                  in the matrix represents the word embedding lookup.
        :raises ValueError: If the mex_length argument is equal to or less
                            than 0. Or if the calculated max_length is 0.
        '''

        if max_length == 0:
            raise ValueError('The max length of a sequence cannot be zero')
        elif max_length < -1:
            raise ValueError('The max length has to be either -1 or above '
                             f'zero not {max_length}')

        # Process the text into integers based on the embeddings given
        all_sequence_data = []
        max_sequence = 0
        for text in texts:
            sequence_data = []
            tokens = self.tokeniser(text)
            for token in tokens:
                if self.lower:
                    token = token.lower()
                # If the token does not exist it should lookup the unknown
                # word vector
                sequence_data.append(self.embeddings.word2index[token])
            sequence_length = len(sequence_data)
            if sequence_length > max_sequence:
                max_sequence = sequence_length
            all_sequence_data.append(sequence_data)
        if max_sequence == 0:
            raise ValueError('The max sequence length is 0 suggesting no '
                             'data was provided for training or testing')
        # Pad the sequences
        # If max pad size is set and training the model set the
        # test_pad_size to max sequence length
        if max_length == -1:
            max_length = max_sequence
        return (max_length,
                preprocessing.sequence.pad_sequences(all_sequence_data,
                                                     maxlen=max_length,
                                                     dtype='int32',
                                                     value=0,
                                                     padding=padding,
                                                     truncating=truncate))

    def fit(self, X: np.ndarray, y: np.ndarray,
            validation_data: Tuple[np.ndarray, np.ndarray],
            verbose: int = 0,
            continue_training: bool = False) -> 'keras.callbacks.History':
        '''
        Fit the model according to the given training and validation data.

        :param X: Training samples matrix, shape = [n_samples, n_features]
        :param y: Training targets, shape = [n_samples]
        :param validation_data: Tuple of `(x_val, y_val)`. Used to evaluate the
                                model at each epoch. Will not be trained on
                                this data.
        :param verbose: 0 = silent, 1 = progress
        :param continue_training: Whether the model that has already been
                                  trained should be trained further.
        :return: A record of training loss values and metrics values at
                 successive epochs, as well as validation loss values and
                 validation metrics values.
        '''
        X_val, y_val = validation_data
        if sum(y_val < 0) or sum(y < 0):
            raise ValueError('The class labels have to be greater than 0')
        X, X_val = self.create_training_text(X, X_val)
        if isinstance(X, tuple):
            X = list(X)
            X_val = list(X_val)

        y, y_val = self.create_training_y(y, y_val)
        num_classes = y.shape[1]
        if verbose:
            print(f'Number of classes in the data {num_classes}')

        if not continue_training:
            self.fitted = False
            self._to_be_reproducible(self.reproducible)
            self.model = self.keras_model(num_classes)
        elif self.fitted and not continue_training:
            raise ValueError('The model is already fitted')

        model = self.model
        if not continue_training:
            model.compile(optimizer=self.optimiser(**self.optimiser_params),
                          metrics=['accuracy'],
                          loss='categorical_crossentropy')

        with tempfile.NamedTemporaryFile() as weight_file:
            # Set up the callbacks
            model_checkpoint = ModelCheckpoint(weight_file.name,
                                               monitor='val_loss',
                                               save_best_only=True,
                                               save_weights_only=True,
                                               mode='min')
            early_stopping = EarlyStopping(monitor='val_loss', mode='min',
                                           patience=self.patience)
            callbacks = [early_stopping, model_checkpoint]
            history = model.fit(X, y, validation_data=(X_val, y_val),
                                epochs=self.epochs, callbacks=callbacks,
                                verbose=verbose, batch_size=self.batch_size)
            # Load the best model from the saved weight file
            model.load_weights(weight_file.name)
        self.model = model
        self.fitted = True
        return history

    def probabilities(self, X: np.ndarray) -> np.ndarray:
        '''
        The probability of each class label for all samples in X.

        :param X: Test samples matrix, shape = [n_samples, n_features]]
        :return: Probability of each class label for all samples, shape = \
        [n_samples, n_classes]
        '''

        if self.fitted is False:
            raise ValueError('The model has not been fitted please run the '
                             '`fit` method.')
        # Convert from a sequence of dictionaries into texts and then integers
        # that represent the tokens in the text within the embedding space.
        sequence_test_data = self._pre_process(X, training=False)
        predicted_values = self.model.predict(sequence_test_data)
        return predicted_values

    def predict(self, X: np.ndarray) -> np.ndarray:
        '''
        Predict class labels for samples in X.

        :param X: Test samples matrix, shape = [n_samples, n_features]
        :return: Predicted class label per sample, shape = [n_samples]
        '''

        return np.argmax(self.probabilities(X), axis=1)

    @staticmethod
    def save(model: 'bella.models.base.KerasModel', save_fp: Path) -> None:
        '''
        Given a Keras Model, mode, path to the folder to save too, and a name
        to save the files it will save the data to restore the model.

        :param model: The machine learning model instance to be saved.
        :param save_fp: File path of the location that the model is to be
                        saved.
        :return: Nothing.
        :raises ValueError: If the model has not been fitted or if the model
                            is not of type
                            :py:class:`bella.models.base.KerasModel`
        '''

        if not isinstance(model, KerasModel):
            raise ValueError('The model parameter has to be of type '
                             f'KearsModel not {type(model)}')
        if model.fitted:
            model_fp = save_fp.with_suffix('.h5')
            model.model.save(model_fp)

            attributes_fp = save_fp.with_suffix('.pkl')
            with attributes_fp.open('wb') as attributes_file:
                # optimiser cannot be pickled
                attributes = model.model_parameters()
                del attributes['class_params']['optimiser']
                pickle.dump(attributes, attributes_file)
        else:
            raise ValueError(f'The model {str(model)} has not been fitted. '
                             'This can be done by using the `fit` method')

    @staticmethod
    def load(load_fp: Path) -> 'bella.models.base.KerasModel':
        '''
        Loads an instance of this class from a file.

        :param load_fp: File path of the location that the model was saved to.
        :return: self
        '''

        model_fp = str(load_fp.with_suffix('.h5').resolve())
        attributes_fp = load_fp.with_suffix('.pkl')
        with attributes_fp.open('rb') as attributes_file:
            attributes = pickle.load(attributes_file)
        # optimiser has to be recovered as it could not be pickled in the
        # model parameters
        keras_model = keras.models.load_model(model_fp)
        attributes['class_params']['optimiser'] = keras_model.optimizer
        model_class = attributes.pop('class')
        model = model_class(**attributes['class_params'])
        for name, class_attr in attributes['class_attrs'].items():
            setattr(model, name, class_attr)
        model.model = keras_model
        model.fitted = True
        return model

    @staticmethod
    def evaluate_parameter(model: 'bella.models.base.KerasModel',
                           train: Tuple[np.ndarray, np.ndarray],
                           val: Tuple[np.ndarray, np.ndarray],
                           test: np.ndarray, parameter_name: str,
                           parameter: Any) -> Tuple[Any, np.ndarray]:
        '''
        Given a model will set the `parameter_name` to `parameter` fit the
        model and return the a Tuple of parameter changed and predictions of
        the model on the test data, using the train and validation data for
        fitting.

        :param model: KerasModel instance
        :param train: Tuple of `(X_train, y_train)`. Used to fit the model.
        :param val: Tuple of `(X_val, y_val)`. Used to evaluate the
                    model at each epoch. Will not be trained on
                    this data.
        :param test: `X_test` data to predict on.
        :param parameter_name: Name of the parameter to change e.g. optimiser
        :param parameter: value to assign to the parameter e.g.
                          :py:class:`keras.optimizers.RMSprop`
        :return: A tuple of (parameter value, predictions)
        '''

        setattr(model, parameter_name, parameter)
        model.fit(train[0], train[1], val)
        predictions = model.predict(test)
        return (parameter, predictions)

    @staticmethod
    def evaluate_parameters(model: 'bella.models.base.KerasModel',
                            train: Tuple[np.ndarray, np.ndarray],
                            val: Tuple[np.ndarray, np.ndarray],
                            test: np.ndarray, parameter_name: str,
                            parameters: List[Any], n_jobs: int
                            ) -> List[Tuple[Any, np.ndarray]]:
        '''
        Performs :py:func:`bella.models.base.KerasModel.evaluate_parameter` on
        one `parameter_name` but with multiple parameter values.

        This is useful if you would like to know the affect of changing the
        values of a parameter. It can also perform the task in a
        multiprocessing manner if `n_jobs` > 1.

        :param model: :py:class:`bella.models.base.KerasModel` instance
        :param train: Tuple of `(X_train, y_train)`. Used to fit the model.
        :param val: Tuple of `(X_val, y_val)`. Used to evaluate the
                    model at each epoch. Will not be trained on
                    this data.
        :param test: `X_test` data to predict on.
        :param parameter_name: Name of the parameter to change e.g. optimiser
        :param parameters: A list of values to assign to the parameter e.g.
                           [:py:class:`keras.optimizers.RMSprop`]
        :param n_jobs: Number of cpus to use for multiprocessing if 1 then
                       will not multiprocess.
        :return: A list of tuples of (parameter value, predictions)
        '''
        func_args = ((model, train, val, test, parameter_name, parameter)
                     for parameter in parameters)
        if n_jobs == 1:
            return [KerasModel.evaluate_parameter(*args)
                    for args in func_args]
        with Pool(n_jobs) as pool:
            return pool.starmap(KerasModel.evaluate_parameter, func_args)

    @staticmethod
    def _to_be_reproducible(reproducible: Union[int, None]) -> None:
        '''
        To make the method reproducible or not. If it is not needed then
        we can use all the python threads.

        :param reproducible: If int is provided this int is used as the seed
                             values. Else None should be given if it is not
                             to be reproducible.
        '''
        if reproducible is not None:
            os.environ['PYTHONHASHSEED'] = '0'
            np.random.seed(reproducible)
            rn.seed(reproducible)
            # Forces tensorflow to use only one thread
            session_conf = tf.ConfigProto(intra_op_parallelism_threads=1,
                                          inter_op_parallelism_threads=1)
            tf.set_random_seed(reproducible)

            sess = tf.Session(graph=tf.get_default_graph(),
                              config=session_conf)
            keras.backend.set_session(sess)
        else:
            np.random.seed(None)
            rn.seed(np.random.randint(0, 1000))
            tf.set_random_seed(np.random.randint(0, 1000))

    @property
    def tokeniser(self) -> Callable[[str], List[str]]:
        '''
        tokeniser attribute

        :return: The tokeniser used in the model
        '''

        return self._tokeniser

    @tokeniser.setter
    def tokeniser(self, value: Callable[[str], List[str]]) -> None:
        '''
        Sets the tokeniser attribute

        :param value: The value to assign to the tokeniser attribute
        '''

        self.fitted = False
        self._tokeniser = value

    @property
    def embeddings(self) -> 'bella.word_vectors.WordVectors':
        '''
        embeddings attribute

        :return: The embeddings used in the model
        '''

        return self._embeddings

    @embeddings.setter
    def embeddings(self, value: 'bella.word_vectors.WordVectors') -> None:
        '''
        Sets the embeddings attribute

        :param value: The value to assign to the embeddings attribute
        '''

        self.fitted = False
        self._embeddings = value

    @property
    def lower(self) -> bool:
        '''
        lower attribute

        :return: The lower used in the model
        '''

        return self._lower

    @lower.setter
    def lower(self, value: bool) -> None:
        '''
        Sets the lower attribute

        :param value: The value to assign to the lower attribute
        '''

        self.fitted = False
        self._lower = value

    @property
    def reproducible(self) -> Union[int, None]:
        '''
        reproducible attribute

        :return: The reproducible used in the model
        '''

        return self._reproducible

    @reproducible.setter
    def reproducible(self, value: Union[int, None]) -> None:
        '''
        Sets the reproducible attribute

        :param value: The value to assign to the reproducible attribute
        '''

        self.fitted = False
        self._reproducible = value

    @property
    def patience(self) -> int:
        '''
        patience attribute

        :return: The patience used in the model
        '''

        return self._patience

    @patience.setter
    def patience(self, value: int) -> None:
        '''
        Sets the patience attribute

        :param value: The value to assign to the patience attribute
        '''

        self.fitted = False
        self._patience = value

    @property
    def batch_size(self) -> int:
        '''
        batch_size attribute

        :return: The batch_size used in the model
        '''

        return self._batch_size

    @batch_size.setter
    def batch_size(self, value: int) -> None:
        '''
        Sets the batch_size attribute

        :param value: The value to assign to the batch_size attribute
        '''

        self.fitted = False
        self._batch_size = value

    @property
    def epochs(self) -> int:
        '''
        epochs attribute

        :return: The epochs used in the model
        '''

        return self._epochs

    @epochs.setter
    def epochs(self, value: int) -> None:
        '''
        Sets the epochs attribute

        :param value: The value to assign to the epochs attribute
        '''

        self.fitted = False
        self._epochs = value

    @property
    def optimiser(self) -> 'keras.optimizers.Optimizer':
        '''
        optimiser attribute

        :return: The optimiser used in the model
        '''

        return self._optimiser

    @optimiser.setter
    def optimiser(self, value: 'keras.optimizers.Optimizer') -> None:
        '''
        Sets the optimiser attribute

        :param value: The value to assign to the optimiser attribute
        '''

        self.fitted = False
        self._optimiser = value

    @property
    def optimiser_params(self) -> Union[Dict[str, Any], None]:
        '''
        optimiser_params attribute

        :return: The optimiser_params used in the model
        '''

        return self._optimiser_params

    @optimiser_params.setter
    def optimiser_params(self, value: Union[Dict[str, Any], None]) -> None:
        '''
        Sets the optimiser_params attribute

        :param value: The value to assign to the optimiser_params attribute
        '''

        self.fitted = False
        self._optimiser_params = value


class SKLearnModel(BaseModel):
    '''
    Concrete class that is designed to be used as the base class for all
    machine learning models that are based on the
    `scikit learn library <http://scikit-learn.org/stable/>`_.

    At the moment expects all of the machine learning models to use a
    `SVM <http://scikit-learn.org/0.19/modules/cla\
    sses.html#module-sklearn.svm>`_ as their classifier. This is due to
    assuming the model will have the method
    :py:meth:`sklearn.svm.SVC.decision_function` to get `probabilities`.

    **NOTE** each time the *model_parameters* are set it resets the model
    i.e. the *fitted* attribute is :py:class:`False`

    Attributes:

    1. model -- Machine learning model. Expects it to be a
       :py:class:`sklearn.pipeline.Pipeline` instance.
    2. fitted -- If the machine learning model has been fitted (default False)
    3. model_parameters -- The parameters that are set in the machine
       learning model. E.g. Parameter could be the tokeniser used.

    Abstract Class Methods:

    1. get_parameters -- Transform the given parameters into a dictonary
       that is accepted as model parameters.
    2. get_cv_parameters -- Transform the given parameters into a list of
       dictonaries that is accepted as `param_grid` parameter in
       :py:class:`sklearn.model_selection.GridSearchCV`
    3. normalise_parameter_names -- Converts the output of 
       :py:meth:`get_parameters` into a dictionary that can be used as input 
       into :py:meth:`get_parameters`. This is required so that the 
       :py:func:`evaluate_parameters` can work this class.

    Methods:

    1. fit -- Fit the model according to the given training data.
    2. predict -- Predict class labels for samples in X.
    3. probabilities -- The probability of each class label for all samples
       in X.
    4. __repr__ -- Name of the machine learning model.

    Functions:

    1. save -- Given a instance of this class will save it to a file.
    2. load -- Loads an instance of this class from a file.
    3. evaluate_parameter -- fit and predict given training, validation and
       test data the given model when the given parameter is changed on the
       model.
    4. evaluate_parameters -- same as evaluate_parameter however it
       evaluates over many parameter values for the same parameter.
    5. grid_search_model -- Given a model class it will perform a Grid Search
       over the parameters you give to the models
       :py:func:`bella.models.base.SKLearnModel.get_cv_parameters` function
       via the keyword arguments. Returns a pandas dataframe representation of
       the grid search results.
    6. get_grid_score -- Given the return of the :py:func:`grid_search_model`
       will return the grid scores as a List of the mean test accuracy result.
    7. models_best_parameter -- Given a list of models and their base model
       arguments, it will find the best parameter value out of the values
       given for that parameter while keeping the base model arguments
       constant for each model.

    Abstract Functions:

    1. Pipeline -- Machine Learning model that is used as the base template
       for the model attribute. Expects it to be a
       :py:class:`sklearn.pipeline.Pipeline` instance.
    '''

    def __init__(self, *args, **kwargs) -> None:

        self.model = self.pipeline()
        self.fitted = False
        self._model_parameters = self.get_parameters(*args, **kwargs)
        self.model.set_params(**self._model_parameters)

    def fit(self, X: np.ndarray, y: np.ndarray):
        '''
        Fit the model according to the given training data.

        :param X: Training samples matrix, shape = [n_samples, n_features]
        :param y: Training targets, shape = [n_samples]
        :return: The `model` attribute will now be trained.
        '''

        self.model.fit(X, y)
        self.fitted = True

    def predict(self, X: np.ndarray):
        '''
        Predict class labels for samples in X.

        :param X: Test samples matrix, shape = [n_samples, n_features]
        :return: Predicted class label per sample, shape = [n_samples]
        :raises ValueError: If the model has not been fitted
        '''

        if self.fitted:
            return self.model.predict(X)
        raise ValueError(f'The model {str(self)} has not been fitted. '
                         'This can be done by using the `fit` method')

    def probabilities(self, X: np.ndarray):
        '''
        The probability of each class label for all samples in X.

        :param X: Test samples matrix, shape = [n_samples, n_features]]
        :return: Probability of each class label for all samples, shape =
                 [n_samples, n_classes]
        :raises ValueError: If the model has not been fitted
        '''

        if self.fitted:
            return self.model.decision_function(X)
        raise ValueError(f'The model {str(self)} has not been fitted. '
                         'This can be done by using the `fit` method')

    @property
    def model_parameters(self) -> Dict[str, Any]:
        '''
        The parameters that are set in the machine learning model. E.g.
        Parameter could be the tokeniser used.

        :return: parameters of the machine learning model
        '''
        return self._model_parameters

    @model_parameters.setter
    def model_parameters(self, value: Dict[str, Any]) -> None:
        '''
        Set the parameters of the machine learning model.

        :param value: The new parameters of the machine learning model
        '''
        self._model_parameters = self.get_parameters(**value)
        self.model.set_params(**self._model_parameters)
        self.fitted = False

    @staticmethod
    def save(model: 'bella.models.base.SKLearnModel',
             save_fp: Path, compress: int = 0) -> None:
        '''
        Given an instance of this class will save it to a file.

        :param model: The machine learning model instance to be saved.
        :param save_fp: File path of the location that the model is to be
                        saved to.
        :param compress: Optional (default 0). Level of compression 0 is no
                         compression and 9 is the most compressed. The more
                         compressed the lower the read/write time.
        :return: Nothing.
        :raises ValueError: If the model has not been fitted or if the model
                            is not of type
                            :py:class:`bella.models.base.SKLearn`
        '''

        if not isinstance(model, SKLearnModel):
            raise ValueError('The model parameter has to be of type '
                             f'SKLearnModel not {type(model)}')
        if model.fitted:
            joblib.dump(model, save_fp, compress=compress)
        else:
            raise ValueError(f'The model {str(model)} has not been fitted. '
                             'This can be done by using the `fit` method')

    @staticmethod
    def load(load_fp: Path) -> 'bella.models.base.SKLearnModel':
        '''
        Loads an instance of this class from a file.

        :param load_fp: File path of the location that the model was saved to.
        :return: self
        '''

        return joblib.load(load_fp)

    @staticmethod
    def evaluate_parameter(model: 'bella.models.base.SKLearnModel',
                           train: Tuple[np.ndarray, np.ndarray],
                           val: None,
                           test: np.ndarray, parameter_name: str,
                           parameter: Any) -> Tuple[Any, np.ndarray]:
        '''
        Given a model will set the `parameter_name` to `parameter` fit the
        model and return the a Tuple of parameter changed and predictions of
        the model on the test data, using the train and validation data for
        fitting.

        :param model: :py:class:`bella.models.base.SKLearn` instance
        :param train: Tuple of `(X_train, y_train)`. Used to fit the model.
        :param val: Use None. This is only kept to keep the API clean.
        :param test: `X_test` data to predict on.
        :param parameter_name: Name of the parameter to change
                               e.g. word_vectors
        :param parameter: value to assign to the parameter e.g.
                          :py:class:`bella.word_vectors.SSWE`
        :return: A tuple of (parameter value, predictions)
        '''

        original_parameters = model._model_parameters
        original_parameters = model.normalise_parameter_names(original_parameters)
        original_parameters[parameter_name] = parameter
        model.model_parameters = original_parameters
        model.fit(train[0], train[1])
        predictions = model.predict(test)
        return (parameter, predictions)

    @staticmethod
    def evaluate_parameters(model: 'bella.models.base.SKLearnModel',
                            train: Tuple[np.ndarray, np.ndarray],
                            val: None,
                            test: np.ndarray, parameter_name: str,
                            parameters: List[Any], n_jobs: int
                            ) -> List[Tuple[Any, np.ndarray]]:
        '''
        Performs :py:func:`bella.models.base.KerasModel.evaluate_parameter` on
        one `parameter_name` but with multiple parameter values.

        This is useful if you would like to know the affect of changing the
        values of a parameter. It can also perform the task in a
        multiprocessing manner if `n_jobs` > 1.

        :param model: :py:class:`bella.models.base.SKLearn` instance
        :param train: Tuple of `(X_train, y_train)`. Used to fit the model.
        :param val: Use None. This is only kept to keep the API clean.
        :param test: `X_test` data to predict on.
        :param parameter_name: Name of the parameter to change e.g.
                               word_vectors
        :param parameters: A list of values to assign to the parameter e.g.
                           [:py:class:`bella.word_vectors.SSWE`]
        :param n_jobs: Number of cpus to use for multiprocessing if 1 then
                       will not multiprocess.
        :return: A list of tuples of (parameter value, predictions)
        '''
        func_args = ((model, train, val, test, parameter_name, parameter)
                     for parameter in parameters)
        if n_jobs == 1:
            return [SKLearnModel.evaluate_parameter(*args)
                    for args in func_args]
        with Pool(n_jobs) as pool:
            return pool.starmap(SKLearnModel.evaluate_parameter, func_args)

    @staticmethod
    def grid_search_model(model: 'bella.models.base.SKLearnModel',
                          X: np.ndarray, y: np.ndarray, n_cpus: int = 1,
                          num_folds: int = 5, **kwargs) -> pd.DataFrame:
        '''
        Given a model class it will perform a Grid Search over the parameters
        you give to the models :py:func:`bella.models.base.SKLearnModel\
        .get_cv_parameters` function via the keyword arguments. Returns a
        pandas dataframe representation of the grid search results.

        :param model: The class of the model to use not an instance of the
                      model.
        :param X: Training samples matrix, shape = [n_samples, n_features]
        :param y: Training targets, shape = [n_samples]
        :param n_cpus: Number of estimators to fit in parallel. Default 1.
        :param num_folds: Number of Stratified cross validation folds.
                          Default 5.
        :param kwargs: Keyword arguments to give to the models
                       :py:func:`bella.models.base.SKLearnModel\
                       .get_cv_parameters` function.
        :return: Pandas dataframe representation of the grid search results.
        '''
        stratified_folds = StratifiedKFold(num_folds)
        grid_params = model.get_cv_parameters(**kwargs)
        grid_model = GridSearchCV(model.pipeline(), grid_params,
                                  cv=stratified_folds, n_jobs=n_cpus,
                                  return_train_score=False)
        grid_model.fit(X, y)
        return pd.DataFrame(grid_model.cv_results_)

    @staticmethod
    def get_grid_score(grid_scores: pd.DataFrame,
                       associated_param: Union[None, str] = None
                       ) -> Union[List[float], List[Tuple[float, str]]]:
        '''
        Given the return of the :py:func:`grid_search_model` will return
        the grid scores as a List of the mean test accuracy result.

        :param grid_scores: Return of the :py:func:`grid_search_model`
        :param associated_param: Optional. The name of the parameter you want
                                 to associate to the score. E.g. lexicon as you
                                 have grid searched over different lexicons and
                                 you want the return to be associated with the
                                 lexicon name e.g. [(0.68, 'MPQA),
                                 (0.70, 'NRC')]
        :return: A list of test scores from the grid search and if
                 associated_param is not None a list of scores and parameter
                 names.
        '''
        extracted_scores = grid_scores['mean_test_score'].astype(float)
        extracted_scores = extracted_scores.round(4) * 100
        extracted_scores = extracted_scores.tolist()
        if associated_param is not None:
            if associated_param not in grid_scores:
                for column_name in grid_scores.columns:
                    if associated_param in column_name:
                        associated_param = column_name
            associated_param = grid_scores[associated_param]
            associated_param = associated_param.apply(str).tolist()
            extracted_scores = list(zip(extracted_scores, associated_param))
        return extracted_scores

    @staticmethod
    def models_best_parameter(models_kwargs: List[Tuple['bella.models.base.SKLearnModel',
                                                        Dict[str, Any]]],
                              param_name: str, param_values: List[Any],
                              X: List[Any], y: np.ndarray, n_cpus: int = 1,
                              num_folds: int = 5
                              ) -> Dict['bella.models.base.SKLearnModel', str]:
        '''
        Given a list of models and their base model arguments, it will
        find the best parameter value out of the values given for that
        parameter while keeping the base model arguments constant for
        each model.

        This essentially performs 5 fold cross validation grid search
        for the one parameter given, across all models given.

        :param models_kwargs: A list of tuples where each tuple contains
                              a model and the models keyword arguments to
                              give to its `get_cv_parameters` method. These
                              arguments are the models standard arguments
                              that are not to be changed.
        :param param_name: Name of the parameter to be changed. This name
                           has to be the name of the keyword argument in
                           the models `get_cv_parameters` method.
        :param param_values: The different values to assign to the param_name
                             argument.
        :param X: The training samples.
        :param y: The training target samples.
        :return: A dictionary of model and the name of the best parameter.
        '''
        model_best_param = {}
        for model, model_kwargs in models_kwargs:
            temp_model_kwargs = {**model_kwargs, param_name: param_values}
            grid_results = model.grid_search_model(model, X, y, n_cpus=n_cpus,
                                                   num_folds=num_folds,
                                                   **temp_model_kwargs)
            param_scores = model.get_grid_score(grid_results, param_name)
            param_scores = sorted(param_scores, key=lambda x: x[1],
                                  reverse=True)
            best_param = sorted(param_scores, key=lambda x: x[0])[-1][1]
            model_best_param[model] = best_param
        return model_best_param

    @classmethod
    @abstractmethod
    def normalise_parameter_names(cls, parameter_dict: Dict[str, Any]
                                  ) -> Dict[str, Any]:
        '''
        Converts the output of :py:meth:`get_parameters` into a dictionary that 
        can be used as input into :py:meth:`get_parameters`.

        :returns: A dictonary that can be used as keyword arguments into the 
                  :py:meth:`get_parameters` method
        '''

        pass

    @classmethod
    @abstractmethod
    def get_parameters(cls) -> Dict[str, Any]:
        '''
        Transform the given parameters into a dictonary that is accepted as
        model parameters
        '''

        pass

    @classmethod
    @abstractmethod
    def get_cv_parameters(cls) -> List[Dict[str, List[Any]]]:
        '''
        Transform the given parameters into a list of dictonaries that is
        accepted as `param_grid` parameter in
        :py:class:`sklearn.model_selection.GridSearchCV`
        '''

        pass

    @staticmethod
    def _add_to_params_dict(params_dict: Dict[str, Any], keys: List[str],
                            value: Any) -> Dict[str, Any]:
        '''
        Given a dictionary it adds the value to each key in the list of keys
        into the dictionary. Returns the updated dictionary.

        Normally used in subclasses :py:meth:`get_parameters`

        :param params_dict: Dictionary to be updated
        :param keys: list of keys
        :param value: value to be added to each key in the list of keys.
        :returns: The dictionary updated
        '''

        if not isinstance(keys, list):
            raise ValueError('The keys parameter has to be of type list and '
                             f'not {type(keys)}')
        for key in keys:
            params_dict[key] = value
        return params_dict

    @staticmethod
    def _add_to_params(params_list: Union[List[Dict[str, List[Any]]], List],
                       to_add: List[Any],
                       to_add_names: List[str]) -> List[Dict[str, List[Any]]]:
        '''
        Used to add parameters that are stated multiple times in the same
        pipeline that must have the same value.

        Therefore to add them you have to copy the current parameter
        list N amount of times where N is the length of the to_add list.
        Returns the updated parameter list. Method to add parameters that
        are set in multiple parts of the pipeline but should contain the
        same value.

        Normally used in subclasses :py:meth:`get_cv_parameters`

        :params_list: A list of dicts where each dict contains parameters and
                      corresponding values that are to be searched for. Can be
                      an empty List.
        :param to_add: List of values that are to be added to the search space.
        :param to_add_names: List of names that are associated to the values.
        :returns: The updated params_list
        :raises TypeError: If any of the arguments are not of type
                           :py:class:`List`
        '''
        # Check the type of the argument
        if not isinstance(params_list, list):
            raise TypeError(f'params_list: {params_list}\nShould be of type '
                            f'list not {type(params_list)}')
        if not isinstance(to_add_names, list):
            raise TypeError(f'to_add_names: {to_add_names}\nShould be of type '
                            f'list not {type(to_add_names)}')
        param_name = to_add_names[0]
        if len(to_add_names) > 1:
            param_name = ''.join(param_name.split('__')[:-1])
        if not isinstance(to_add, list):
            raise TypeError('If using get_cv_parameters this is due to '
                            f'parameter {param_name} not being of type list.'
                            f'\nto_add: {to_add} should be of '
                            f'type List not {type(to_add)}.')

        num_params = len(params_list)
        num_to_add = len(to_add)
        new_param_list = []
        # Catch the case that params_list was originally empty
        if num_params == 0:
            for _ in range(num_to_add):
                new_param_list.append([defaultdict(list)])
        else:
            for _ in range(num_to_add):
                new_param_list.append(copy.deepcopy(params_list))

        for index, param in enumerate(to_add):
            for param_name in to_add_names:
                for sub_list in new_param_list[index]:
                    sub_list[param_name].append(param)
        params_list = [param_dict for sub_list in new_param_list
                       for param_dict in sub_list]
        return params_list

    @staticmethod
    def _add_to_all_params(params_list: List[Dict[str, List[Any]]],
                           param_name: str, param_value: List[Any]
                           ) -> List[Dict[str, List[Any]]]:
        '''
        Used to add param_name and its associated param_value to each
        dictionary of parameters in the params_list.

        Normally used in subclasses :py:meth:`get_cv_parameters`

        :param params_list: A list of dicts where each dict contains
                            parameters and corresponding values that are to be
                            searched for.
        :param param_name: The name associated to the parameter value to be
                           added to the params_list.
        :param param_value: The list of values associated to the param_name
                            that are added to the params_list.
        :returns: The updated params_list
        :raises TypeError: If the param_value is not of type :py:class:`List`
        '''

        if not isinstance(param_value, list):
            raise TypeError(f'{param_name} should be of type list not '
                            f'{type(param_value)}')
        for param_dict in params_list:
            param_dict[param_name] = param_value
        return params_list

    @staticmethod
    @abstractmethod
    def pipeline() -> 'sklearn.pipeline.Pipeline':
        '''
        Machine Learning model that is used as the base template for the model
        attribute.

        :returns: The template machine learning model
        '''
        pass



'''
Contains the following neural pooling functions:

1. min
2. max
3. avg

Which are from
`Tang et al <https://aclanthology.coli.uni-saarland.de/papers/P14-1146/p14-1146>`_.

and the following pooling functions:

4. prod
5. std

Which are from
`Vo and Zhang <https://www.ijcai.org/Proceedings/15/Papers/194.pdf>`_.

and finally the following pooling function:

6. median

From `Bo Wang et al.
<https://aclanthology.coli.uni-saarland.de/papers/E17-1046/e17-1046>`_

All the functions are applied over the columns and not the rows e.g. matrix
of (m, n) size and apply mean it will return a vector of (1, n). Therefore by
default all of the vectors returned are row vectors but if transpose is True
then column vectors are returned.
'''
from functools import wraps

import numpy as np

def inf_nan_check(neural_func):
    '''
    Contains decorator function that converts any inf or NAN value to a real
    number to avoid any potential problems with inf and NAN's latter on in the
    processing chain.

    Inf conversion - Converts it to the max (min) value of the numpy array/matrix
    dtype based on it being positive (negative) value.

    NAN conversion - based on the following
    `post <https://stackoverflow.com/questions/25506281/what-are-all-the-possib\
    le-calculations-that-could-cause-a-nan-in-python>`_ about how NAN's occur.
    It converts NAN's to zeros as the majority of the operation should equal
    zero or are close to zero. This is a rough approximation but it should not
    affect that many numbers.
    '''

    @wraps(neural_func)
    def func_wrapper(matrix, **kwargs):
        '''
        :param matrix: Numpy array/matrix that could contain NAN or inf values.
        :param transpose: If to convert the column vector into row vector
        :type matrix: np.ndarray
        :type transpose: bool
        :returns: The numpy array/matrix with NAN and inf values converted to \
        real values.
        :rtype: np.ndarray
        '''

        matrix = neural_func(matrix, **kwargs)
        if not issubclass(matrix.dtype.type, np.floating):
            raise TypeError('Only accept floating value word embeddings not '\
                            '{}'.format(matrix.dtype.type))
        # Convert all NAN values to zero
        if np.any(np.isnan(matrix)):
            matrix[np.where(np.isnan(matrix))] = 0
        # Find any value that is greater than half the min and max values and
        # convert them to half the min or max value respectively. This is
        # done to ensure that range can be done without overflow exception
        dtype_info = np.finfo(matrix.dtype)
        min_value = dtype_info.min / 2
        max_value = dtype_info.max / 2
        if np.any(matrix[matrix < min_value]) or np.any(matrix[matrix > max_value]):
            matrix[matrix < min_value] = min_value
            matrix[matrix > max_value] = max_value

        return matrix

    return func_wrapper

def matrix_checking(neural_func):
    '''
    Contains decorator function to check argument compbatbility and the
    decorated functions return. The functions decorated are the neural functions
    which are:

    1. :py:func:`bella.neural_pooling.matrix_min`
    2. :py:func:`bella.neural_pooling.matrix_max`
    3. :py:func:`bella.neural_pooling.matrix_avg`
    '''
    @wraps(neural_func)
    def func_wrapper(matrix, transpose=False):
        '''
        Checks the matrix is of the correct type and that the return matrix
        is of the correct size after the neural_func function has been applied to
        the matrix.

        inf values are converted to max (min) value defined by the dtype if
        the value is positive (negative).

        Applies transpose to convert row vectors into column vectors if
        transpose == False

        :param matrix: matrix or vector
        :param transpose: If to convert the column vector into row vector
        :type matrix: np.ndarray
        :type transpose: bool
        :returns: The output of the neural_func function.
        :rtype: np.ndarray
        '''
        # Pre check
        if not isinstance(matrix, np.ndarray):
            raise TypeError('The matrix has to be of type numpy.ndarray and not '\
                            '{}'.format(type(matrix)))
        # Applying the relevant neural pooling function
        reduced_matrix = neural_func(matrix)
        # Post check
        rm_cols = reduced_matrix.shape[0]
        rm_dim = len(reduced_matrix.shape)
        if rm_dim != 1:
            raise ValueError('The returned matrix should be a vector and have '\
                             'a dimension of 1 it is: {}'.format(rm_dim))
        m_columns = matrix.shape[1]
        if rm_cols != m_columns:
            raise ValueError('The number of columns has changed during the pooling'\
                             'func from {} to {}'.format(m_columns, rm_cols))
        if transpose:
            return reduced_matrix.reshape(rm_cols, 1)
        return reduced_matrix.reshape(1, rm_cols)
    return func_wrapper

@inf_nan_check
@matrix_checking
def matrix_min(matrix, **kwargs):
    '''
    :param matrix: matrix or vector
    :param kwargs: Can keywords that are accepted by `matrix_checking` function
    :type matrix: np.ndarray
    :type kwargs: dict
    :returns: The minimum column values in the matrix.
    :rtype: np.ndarray
    '''

    return matrix.min(axis=0)

@inf_nan_check
@matrix_checking
def matrix_max(matrix, **kwargs):
    '''
    :param matrix: matrix or vector
    :param kwargs: Can keywords that are accepted by `matrix_checking` function
    :type matrix: np.ndarray
    :type kwargs: dict
    :returns: The maximum column values in the matrix.
    :rtype: np.ndarray
    '''

    return matrix.max(axis=0)

@inf_nan_check
@matrix_checking
def matrix_avg(matrix, **kwargs):
    '''
    :param matrix: matrix or vector
    :param kwargs: Can keywords that are accepted by `matrix_checking` function
    :type matrix: np.ndarray
    :type kwargs: dict
    :returns: The mean column values in the matrix.
    :rtype: np.ndarray
    '''

    return matrix.mean(axis=0)

@inf_nan_check
@matrix_checking
def matrix_median(matrix, **kwargs):
    '''

    :param matrix: matrix or vector
    :param kwargs: Can keywords that are accepted by `matrix_checking` function
    :type matrix: np.ndarray
    :type kwargs: dict
    :returns: The median column values in the matrix.
    :rtype: np.ndarray
    '''

    return np.median(matrix, axis=0)

@inf_nan_check
@matrix_checking
def matrix_std(matrix, **kwargs):
    '''
    :param matrix: matrix or vector
    :param kwargs: Can keywords that are accepted by `matrix_checking` function
    :type matrix: np.ndarray
    :type kwargs: dict
    :returns: The standard deviation of the column values in the matrix.
    :rtype: np.ndarray
    '''
    return np.std(matrix, axis=0)

@inf_nan_check
@matrix_checking
def matrix_prod(matrix, **kwargs):
    '''
    :param matrix: matrix or vector
    :param kwargs: Can keywords that are accepted by `matrix_checking` function
    :type matrix: np.ndarray
    :type kwargs: dict
    :returns: The product of the column values in the matrix.
    :rtype: np.ndarray
    '''
    return np.prod(matrix, axis=0)



'''
Module contains all of the classes that represent Machine Learning models
that are within `Wang et al. paper \
<https://aclanthology.coli.uni-saarland.de/papers/E17-1046/e17-1046>`_.

1. :py:class:`bella.models.target.TDParseMinus` -- TDParse Minus model
2. :py:class:`bella.models.target.TDParse` -- TDParse model
3. :py:class:`bella.models.tdparse.TDParsePlus` -- TDParse Plus model
'''
from typing import Any, List, Callable, Dict

import sklearn
from sklearn.pipeline import FeatureUnion, Pipeline
from sklearn.preprocessing import MinMaxScaler
from sklearn.svm import LinearSVC

import bella
from bella.tokenisers import ark_twokenize
from bella.neural_pooling import matrix_max, matrix_min, matrix_avg
from bella.neural_pooling import matrix_median, matrix_prod, matrix_std
from bella.scikit_features.context import Context
from bella.scikit_features import syntactic_context
from bella.scikit_features.lexicon_filter import LexiconFilter
from bella.scikit_features.join_context_vectors import JoinContextVectors
from bella.scikit_features.neural_pooling import NeuralPooling
from bella.scikit_features.tokeniser import ContextTokeniser
from bella.scikit_features.word_vector import ContextWordVectors
from bella.models.target import TargetInd


class TDParseMinus(TargetInd):

    @classmethod
    def name(cls) -> str:
        return 'TDParse Minus'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def __init__(self, word_vectors: List['bella.word_vectors.WordVectors'],
                 parser: Any,
                 tokeniser: Callable[[str], List[str]] = ark_twokenize,
                 lower: bool = True, C: float = 0.01,
                 random_state: int = 42,
                 scale: Any = MinMaxScaler()) -> None:
        '''
        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param parser: The dependency parser to be used.
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Whether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        '''
        # Inherit from SKLearnModel __init__ method
        # The parameters here go into the self.get_parameters method
        super(TargetInd, self).__init__(word_vectors, parser, tokeniser, lower,
                                        C, random_state, scale)

    @staticmethod
    def pipeline() -> 'sklearn.pipeline.Pipeline':
        '''
        Machine Learning model that is used as the base template for the model
        attribute.

        :returns: The template machine learning model
        '''
        return Pipeline([
            ('dependency_context', syntactic_context.SyntacticContext()),
            ('contexts', syntactic_context.Context('full')),
            ('tokens', ContextTokeniser()),
            ('word_vectors', ContextWordVectors()),
            ('pool_funcs', FeatureUnion([
                ('max_pipe', Pipeline([
                    ('max', NeuralPooling(matrix_max)),
                    ('join', JoinContextVectors(matrix_median))
                ])),
                ('min_pipe', Pipeline([
                    ('min', NeuralPooling(matrix_min)),
                    ('join', JoinContextVectors(matrix_median))
                ])),
                ('avg_pipe', Pipeline([
                    ('avg', NeuralPooling(matrix_avg)),
                    ('join', JoinContextVectors(matrix_median))
                ])),
                ('prod_pipe', Pipeline([
                    ('min', NeuralPooling(matrix_prod)),
                    ('join', JoinContextVectors(matrix_median))
                ])),
                ('std_pipe', Pipeline([
                    ('min', NeuralPooling(matrix_std)),
                    ('join', JoinContextVectors(matrix_median))
                ]))
            ])),
            ('scale', MinMaxScaler()),
            ('svm', LinearSVC())
        ])

    @classmethod
    def normalise_parameter_names(cls, parameter_dict: Dict[str, Any]
                                  ) -> Dict[str, Any]:
        '''
        Converts the output of :py:meth:`get_parameters` into a dictionary that 
        can be used as input into :py:meth:`get_parameters`.

        :returns: A dictonary that can be used as keyword arguments into the 
                  :py:meth:`get_parameters` method
        '''

        parameter_names = ['word_vectors__', '__tokeniser', '__lower', 'scale',
                           '__C', '__random_state', '__parser']
        name_parameter = {}
        for parameter_name in parameter_names:
            for name, parameter in parameter_dict.items():
                if parameter_name in name:
                    real_parameter_name = parameter_name.strip('_')
                    name_parameter[real_parameter_name] = parameter
        return name_parameter

    @classmethod
    def get_parameters(cls,
                       word_vectors: List['bella.word_vectors.WordVectors'],
                       parser: Any,
                       tokeniser: Callable[[str], List[str]] = ark_twokenize,
                       lower: bool = True, C: float = 0.01,
                       random_state: int = 42,
                       scale: Any = MinMaxScaler()) -> Dict[str, Any]:
        '''
        Transform the given parameters into a dictonary that is accepted as
        model parameters

        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param parser: The dependency parser to be used.
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Whether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        :return: Model parameters
        '''
        params_list = super().get_parameters(word_vectors, tokeniser,
                                             lower, C, random_state, scale)
        params_list = cls._add_to_params_dict(params_list,
                                              cls._get_dependency_context(),
                                              parser)
        return params_list

    @classmethod
    def get_cv_parameters(cls,
                          word_vectors: List[List['bella.word_vectors\
                                                        .WordVectors']],
                          parser: List[Any],
                          tokeniser=[ark_twokenize],
                          lower=[True], C=[0.01], random_state=[42],
                          scale=[MinMaxScaler()]):
        '''
        Transform the given parameters into a list of dictonaries that is
        accepted as `param_grid` parameter in
        :py:class:`sklearn.model_selection.GridSearchCV`

        :param word_vectors: A list of a list of word vectors e.g. [[SSWE()],
                             [SSWE(), GloveCommonCrawl()]].
        :param parser: A list of dependency parser to be used.
        :param tokenisers: A list of tokeniser to be used e.g.
                           :py:meth:`str.split`. Default [ark_twokenize]
        :param lowers: A list of bool values which indicate whether to lower
                       case the input words. Default [True]
        :param C: A list of C values for the :py:class:`sklearn.svm.SVC`
                  estimator that is used in the pipeline. Default [0.01]
        :param random_state: A list of random_state values for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline. Default [42]
        :param scale: List of scale values. The list can include
                      :py:class:`sklearn.preprocessing.MinMaxScaler` type of
                      clases or None if no scaling is to be used. Default
                      [:py:class:`sklearn.preprocessing.MinMaxScaler`]
        :return: Parameters to explore through cross validation
        '''
        params_list = super().get_cv_parameters(word_vectors, tokeniser,
                                                lower, C, random_state, scale)
        # dependency parser
        dep_context = cls._get_dependency_context()[0]
        params_list = cls._add_to_all_params(params_list, dep_context,
                                             parser)
        return params_list

    @staticmethod
    def _get_dependency_context() -> List[str]:
        '''
        Method to be overidden by subclasses as each pipeline will be different
        and will have a different parameter name for where the dependency
        parser process.

        :returns: A list of parameters names where the dependency parser is
                  set in the pipeline
        '''

        return ['dependency_context__parser']


class TDParse(TDParseMinus):

    @classmethod
    def name(cls) -> str:
        return 'TDParse'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def __init__(self, word_vectors: List['bella.word_vectors.WordVectors'],
                 parser: Any,
                 tokeniser: Callable[[str], List[str]] = ark_twokenize,
                 lower: bool = True, C: float = 0.01,
                 random_state: int = 42,
                 scale: Any = MinMaxScaler()) -> None:
        '''
        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param parser: The dependency parser to be used.
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Whether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        '''
        # Inherit from SKLearnModel __init__ method
        # The parameters here go into the self.get_parameters method
        super(TargetInd, self).__init__(word_vectors, parser, tokeniser, lower,
                                        C, random_state, scale)

    @staticmethod
    def pipeline() -> 'sklearn.pipeline.Pipeline':
        '''
        Machine Learning model that is used as the base template for the model
        attribute.

        :returns: The template machine learning model
        '''
        return Pipeline([
            ('union', FeatureUnion([
                ('dependency', Pipeline([
                    ('context', syntactic_context.SyntacticContext()),
                    ('contexts', syntactic_context.Context('full')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('left', Pipeline([
                    ('contexts', Context('left')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('right', Pipeline([
                    ('contexts', Context('right')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('target', Pipeline([
                    ('contexts', Context('target')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ]))
            ])),
            ('scale', MinMaxScaler()),
            ('svm', LinearSVC())
        ])

    @staticmethod
    def _get_word_vector_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the word vectors are set
        in the pipeline.
        '''

        return ['union__dependency__word_vectors__vectors',
                'union__left__word_vectors__vectors',
                'union__right__word_vectors__vectors',
                'union__target__word_vectors__vectors']

    @staticmethod
    def _get_tokeniser_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the tokenisers are set
        in the pipeline.
        '''

        return ['union__dependency__tokens',
                'union__left__tokens',
                'union__right__tokens',
                'union__target__tokens']

    @staticmethod
    def _get_dependency_context() -> List[str]:
        '''
        :returns: A list of parameters names where the dependency parser is
                  set in the pipeline
        '''

        return ['union__dependency__context__parser']


class TDParsePlus(TDParseMinus):

    @classmethod
    def name(cls) -> str:
        return 'TDParsePlus'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def __init__(self, word_vectors: List['bella.word_vectors.WordVectors'],
                 parser: Any, senti_lexicon: 'bella.lexicons.Lexicon',
                 tokeniser: Callable[[str], List[str]] = ark_twokenize,
                 lower: bool = True, C: float = 0.01,
                 random_state: int = 42,
                 scale: Any = MinMaxScaler()) -> None:
        '''
        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param parser: The dependency parser to be used.
        :param senti_lexicon: Sentiment Lexicon to be used for the Left and
                              Right sentiment context (LS and RS).
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Whether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        '''

        # Inherit from SKLearnModel __init__ method
        # The parameters here go into the self.get_parameters method
        super(TargetInd, self).__init__(word_vectors, parser, senti_lexicon,
                                        tokeniser, lower, C, random_state,
                                        scale)

    @staticmethod
    def pipeline() -> 'sklearn.pipeline.Pipeline':
        '''
        Machine Learning model that is used as the base template for the model
        attribute.

        :returns: The template machine learning model
        '''
        return Pipeline([
            ('union', FeatureUnion([
                ('dependency', Pipeline([
                    ('context', syntactic_context.SyntacticContext()),
                    ('contexts', syntactic_context.Context('full')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('left', Pipeline([
                    ('contexts', Context('left')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('right', Pipeline([
                    ('contexts', Context('right')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('left_s', Pipeline([
                    ('contexts', Context('left')),
                    ('tokens', ContextTokeniser()),
                    ('filter', LexiconFilter()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('right_s', Pipeline([
                    ('contexts', Context('right')),
                    ('tokens', ContextTokeniser()),
                    ('filter', LexiconFilter()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('target', Pipeline([
                    ('contexts', Context('target')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ]))
            ])),
            ('scale', MinMaxScaler()),
            ('svm', LinearSVC())
        ])

    @classmethod
    def normalise_parameter_names(cls, parameter_dict: Dict[str, Any]
                                  ) -> Dict[str, Any]:
        '''
        Converts the output of :py:meth:`get_parameters` into a dictionary that 
        can be used as input into :py:meth:`get_parameters`.

        :returns: A dictonary that can be used as keyword arguments into the 
                  :py:meth:`get_parameters` method
        '''

        parameter_names = ['word_vectors__', '__tokeniser', '__lower', 'scale',
                           '__C', '__random_state', '__parser', '__lexicon']
        name_parameter = {}
        for parameter_name in parameter_names:
            for name, parameter in parameter_dict.items():
                if parameter_name in name:
                    real_parameter_name = parameter_name.strip('_')
                    if parameter_name == '__lexicon':
                        real_parameter_name = 'senti_lexicon'
                    name_parameter[real_parameter_name] = parameter
        return name_parameter

    @classmethod
    def get_parameters(cls,
                       word_vectors: List['bella.word_vectors.WordVectors'],
                       parser: Any,
                       senti_lexicon: 'bella.lexicons.Lexicon',
                       tokeniser: Callable[[str], List[str]] = ark_twokenize,
                       lower: bool = True, C: float = 0.01,
                       random_state: int = 42,
                       scale: Any = MinMaxScaler()) -> Dict[str, Any]:
        '''
        Transform the given parameters into a dictonary that is accepted as
        model parameters

        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param parser: The dependency parser to be used.
        :param senti_lexicon: Sentiment Lexicon to be used for the Left and
                              Right sentiment context (LS and RS).
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Whether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is used
                             in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        :return: Model parameters
        '''
        params_dict = super().get_parameters(word_vectors, parser, tokeniser,
                                             lower, C, random_state, scale)
        params_dict = cls._add_to_params_dict(params_dict,
                                              cls._get_word_senti_names(),
                                              senti_lexicon)
        return params_dict

    @classmethod
    def get_cv_parameters(cls,
                          word_vectors: List[List['bella.word_vectors\
                                                        .WordVectors']],
                          parser: List[Any],
                          senti_lexicon: List['bella.lexicons\
                                                    .Lexicon'],
                          tokeniser=[ark_twokenize],
                          lower=[True], C=[0.01], random_state=[42],
                          scale=[MinMaxScaler()]):
        '''
        Transform the given parameters into a list of dictonaries that is
        accepted as `param_grid` parameter in
        :py:class:`sklearn.model_selection.GridSearchCV`

        :param word_vectors: A list of a list of word vectors e.g. [[SSWE()],
                             [SSWE(), GloveCommonCrawl()]].
        :param parser: A list of dependency parser to be used.
        :param senti_lexicon: A list of Sentiment Lexicons to be explored for
                              the Left and Right sentiment context (LS and RS).
        :param tokenisers: A list of tokeniser to be used e.g.
                           :py:meth:`str.split`. Default [ark_twokenize]
        :param lowers: A list of bool values which indicate whether to lower
                       case the input words. Default [True]
        :param C: A list of C values for the :py:class:`sklearn.svm.SVC`
                  estimator that is used in the pipeline. Default [0.01]
        :param random_state: A list of random_state values for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline. Default [42]
        :param scale: List of scale values. The list can include
                      :py:class:`sklearn.preprocessing.MinMaxScaler` type of
                      clases or None if no scaling is to be used. Default
                      [:py:class:`sklearn.preprocessing.MinMaxScaler`]
        :return: Parameters to explore through cross validation
        '''
        params_list = super().get_cv_parameters(word_vectors, parser,
                                                tokeniser, lower, C,
                                                random_state, scale)
        # sentiment lexicon
        params_list = cls._add_to_params(params_list, senti_lexicon,
                                         cls._get_word_senti_names())
        return params_list

    @staticmethod
    def _get_word_vector_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the word vectors are set
        in the pipeline.
        '''

        return ['union__dependency__word_vectors__vectors',
                'union__left__word_vectors__vectors',
                'union__right__word_vectors__vectors',
                'union__target__word_vectors__vectors',
                'union__right_s__word_vectors__vectors',
                'union__left_s__word_vectors__vectors']

    @staticmethod
    def _get_tokeniser_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the tokenisers are set
        in the pipeline.
        '''

        return ['union__dependency__tokens',
                'union__left__tokens',
                'union__right__tokens',
                'union__target__tokens',
                'union__right_s__tokens',
                'union__left_s__tokens']

    @staticmethod
    def _get_dependency_context() -> List[str]:
        '''
        :returns: A list of parameters names where the dependency parser is
                  set in the pipeline
        '''

        return ['union__dependency__context__parser']

    @staticmethod
    def _get_word_senti_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the sentiment lexicons
        are set in the pipeline.
        '''

        return ['union__left_s__filter__lexicon',
                'union__right_s__filter__lexicon']



'''
Functions that parse the annotated data that is being used in this project. The
annotated dataset are the following:

1. `Li Dong <http://goo.gl/5Enpu7>`_ which links to :py:func:`bella.parsers.dong`
2. Semeval parser
'''
import json
import os
import re
import xml.etree.ElementTree as ET
from typing import Tuple
from pathlib import Path

import ftfy

from bella.data_types import Target, TargetCollection

def dong(file_path: Path, **target_collection_kwargs) -> TargetCollection:
    '''
    Given file path to the
    `Li Dong <https://github.com/bluemonk482/tdparse/tree/master/data/lidong>`_
    sentiment data it will parse the data and return it as a list of dictionaries.

    :param file_path: File Path to the annotated data
    :param target_collection_kwargs: Keywords to parse to the TargetCollection 
                                     constructor that is returned.
    :returns: A TargetCollection containing Target instances.
    '''

    file_path = os.path.abspath(file_path)
    if not os.path.isfile(file_path):
        raise FileNotFoundError('This file does not exist {}'.format(file_path))
    file_name, _ = os.path.splitext(os.path.basename(file_path))
    sentiment_range = [-1, 0, 1]

    sentiment_data = TargetCollection(**target_collection_kwargs)
    with open(file_path, 'r') as dong_file:
        sent_dict = {}
        for index, line in enumerate(dong_file):
            divisible = index + 1
            line = line.strip()
            if divisible % 3 == 1:
                sent_dict['text'] = line
            elif divisible % 3 == 2:
                sent_dict['target'] = line
            elif divisible % 3 == 0:
                sentiment = int(line)
                if sentiment not in sentiment_range:
                    raise ValueError('The sentiment has to be one of the '\
                                     'following values {} not {}'\
                                     .format(sentiment_range, sentiment))
                sent_dict['sentiment'] = int(line)
                text = sent_dict['text'].lower()
                target = sent_dict['target'].lower()
                offsets = [match.span() for match in re.finditer(target, text)]
                if len(target.split()) > 1:
                    joined_target = ''.join(target.split())
                    offsets.extend([match.span()
                                    for match in re.finditer(joined_target, text)])
                sent_dict['spans'] = offsets
                sent_id = file_name + str(len(sentiment_data))
                # Sentence ID is the same as the target as there is only one
                # target per sentence
                sent_dict['sentence_id'] = sent_id
                sent_dict['target_id'] = sent_id
                sent_target = Target(**sent_dict)
                sentiment_data.add(sent_target)
                sent_dict = {}
            else:
                raise Exception('Problem')
    return sentiment_data




def _semeval_extract_data(sentences, file_name, conflict=False,
                          sentence_ids_skip=None, **target_collection_kwargs
                          ) -> TargetCollection:
    '''
    :param sentences: A `sentences` named element
    :param file_name: Name of the file being parsed
    :param conflict: Determine if to keep the target data that has a conflict \
    sentiment label.
    :param sentence_ids_skip: IDs of sentences that should be skipped
    :type sentences: xml.etree.ElementTree.Element
    :type file_name: String
    :type conflict: bool. Defailt False
    :type sentence_ids_skip: list. Default None
    :returns: A TargetCollection containing Target instances.
    :rtype: TargetCollection
    '''

    # Converts the sentiment tags from Strings to ints
    sentiment_mapper = {'conflict' : -2, 'negative' : -1,
                        'neutral' : 0, 'positive' : 1}

    def extract_aspect_terms(aspect_terms, sentence_id, 
                             raise_error_no_category=False):
        '''
        :param aspect_terms: An aspectTerms element within the xml tree
        :param sentence_id: Id of the sentence that the aspects came from.
        :type aspect_terms: xml.etree.ElementTree.Element
        :type sentence_id: String
        :returns: A list of dictioanries containg id, span, sentiment and \
        target
        :rtype: list
        '''

        aspect_terms_data = []
        for index, aspect_term in enumerate(aspect_terms):
            aspect_term = aspect_term.attrib
            aspect_term_data = {}
            sentiment = sentiment_mapper[aspect_term['polarity']]
            if sentiment == -2 and not conflict:
                continue
            aspect_id = '{}{}'.format(sentence_id, index)
            aspect_term_data['target_id'] = aspect_id
            if 'term' in aspect_term:
                aspect_term_data['target'] = aspect_term['term']
            elif 'target' in aspect_term:
                aspect_term_data['target'] = aspect_term['target']
            else:
                raise KeyError('There is no `target` attribute in the opinions '\
                               'element {}'.format(aspect_term))
            # Extract the category if it exists
            if 'category' in aspect_term:
                aspect_term_data['category'] = aspect_term['category']
            elif raise_error_no_category:
                raise KeyError('These is no `category` attribute in the '
                               f'opinions element {aspect_term}') 
            aspect_term_data['sentiment'] = sentiment
            aspect_term_data['spans'] = [(int(aspect_term['from']),
                                          int(aspect_term['to']))]
            aspect_term_data['sentence_id'] = sentence_id
            # If the target is NULL then there is no target
            if aspect_term_data['target'] == 'NULL':
                continue
            aspect_terms_data.append(aspect_term_data)
        return aspect_terms_data

    def add_text(aspect_data, text):
        '''
        :param aspect_data: A list of dicts containing `span`, `target` and \
        `sentiment` keys.
        :param text: The text of the sentence that is associated to all of the \
        aspects in the aspect_data list
        :type aspect_data: list
        :type text: String
        :returns: The list of dicts in the aspect_data parameter but with a \
        `text` key with the value that the text parameter contains
        :rtype: list
        '''

        for data in aspect_data:
            data['text'] = text
        return aspect_data

    all_aspect_term_data = TargetCollection(**target_collection_kwargs)
    for sentence in sentences:
        aspect_term_data = None
        text_index = None
        sentence_id = file_name + sentence.attrib['id']
        # Allow the parser to skip certain sentences
        if sentence_ids_skip is not None:
            if sentence.attrib['id'] in sentence_ids_skip:
                continue
        for index, data in enumerate(sentence):
            if data.tag == 'sentence':
                raise Exception(sentence.attrib['id'])
            if data.tag == 'text':
                text_index = index
            elif data.tag == 'aspectTerms' or data.tag == 'Opinions':
                aspect_term_data = extract_aspect_terms(data, sentence_id)
        if aspect_term_data is None:
            continue
        if text_index is None:
            raise ValueError('A semeval sentence should always have text '\
                             'semeval file {} sentence id {}'\
                             .format(file_name, sentence.attrib['id']))
        sentence_text = sentence[text_index].text
        aspect_term_data = add_text(aspect_term_data, sentence_text)
        for aspect in aspect_term_data:
            sent_target = Target(**aspect)
            all_aspect_term_data.add(sent_target)
    return all_aspect_term_data


def semeval_15_16(file_path: Path, sep_16_from_15: bool = False, 
                  sep_15_from_14: bool = False, 
                  raise_error_no_category: bool = False, 
                  **target_collection_kwargs) -> TargetCollection:
    '''
    Parser for the SemEval 2015 and 2016 datasets.

    :param file_path: File path to the semeval 2015/16 data
    :param sep_16_from_15: Ensure that the test sets of semeval 2016 is complete 
                           seperate from the semeval test set of 2015
    :param target_collection_kwargs: Keywords to parse to the TargetCollection 
                                     constructor that is returned.
    :returns: A TargetCollection containing Target instances.
    '''

    file_path = os.path.abspath(file_path)
    file_name, _ = os.path.splitext(os.path.basename(file_path))

    tree = ET.parse(file_path)
    reviews = tree.getroot()
    all_aspect_term_data = []

    ids_to_skip = []
    if sep_16_from_15:
        ids_to_skip = ["en_SnoozeanAMEatery_480032670:4"]
    elif sep_15_from_14:
        ids_to_skip = ['1253117:2',
                       '397331:1']
    if reviews.tag != 'Reviews':
        raise ValueError('The root of all semeval 15/16 xml files should '
                         f'be reviews and not {reviews.tag}')
    for review in reviews:
        review_id = review.attrib['rid']
        for sentences in review:
            review_targets = _semeval_extract_data(sentences, file_name,
                                                   sentence_ids_skip=ids_to_skip)
            all_aspect_term_data.extend(review_targets.data())
    return TargetCollection(all_aspect_term_data, **target_collection_kwargs)

def semeval_14(file_path: Path, conflict: bool = False, 
               **target_collection_kwargs) -> TargetCollection:
    '''
    Parser for the SemEval 2014 datasets.

    :param file_path: File path to the semeval 2014 data
    :param conflict: determine if to include the conflict sentiment value
    :param target_collection_kwargs: Keywords to parse to the TargetCollection 
                                     constructor that is returned.
    :returns: A TargetCollection containing Target instances.
    '''
    file_path = os.path.abspath(file_path)
    file_name, _ = os.path.splitext(os.path.basename(file_path))

    tree = ET.parse(file_path)
    sentences = tree.getroot()
    if sentences.tag != 'sentences':
        raise ValueError('The root of all semeval xml files should '\
                         'be sentences and not {}'\
                         .format(sentences.tag))
    return _semeval_extract_data(sentences, file_name, conflict=conflict,
                                 **target_collection_kwargs)


def election(folder_path: Path, include_dnr: bool = False, 
             include_additional: bool = False
             ) -> Tuple[TargetCollection, TargetCollection]:
    '''
    Data can be downloaded from
    `FigShare <https://figshare.com/articles/EACL_2017_-_Multi-target_\
    UK_election_Twitter_sentiment_corpus/4479563/1>`_

    :param folder_path: Path to the folder containing the data after it has
                        been unziped and all folders within it have been 
                        unziped.
    :param include_dnr: determine if to include the `doesnotapply` label
    :param include_additional: NOTE: This does not work at the moment. 
                               Determine if to parse the additional data.
    :returns: A TargetCollection containing Target instances.
    '''

    sentiment_mapper = {'negative' : -1, 'neutral' : 0, 'positive' : 1}
    folder_path = os.path.abspath(folder_path)
    folder_name, _ = os.path.splitext(os.path.basename(folder_path))

    def get_file_data(folder_dir):
        '''
        :param folder_dir: File path to a folder containing JSON data files \
        where the file names is the datas ID
        :type folder_dir: String
        :returns: A dictionary of IDs as keys and JSON data as values
        :rtype: dict
        '''

        data = {}
        for file_name in os.listdir(folder_dir):
            file_path = os.path.join(folder_dir, file_name)
            tweet_id = file_name.rstrip('.json').lstrip('5')
            with open(file_path, 'r') as file_data:
                data[tweet_id] = json.load(file_data)
        return data

    def parse_tweet(tweet_data, anno_data, tweet_id):

        def get_offsets(entity, tweet_text, target):
            offset_shifts = [0, -1, 1]
            from_offset = entity['offset']
            for offset_shift in offset_shifts:
                from_offset_shift = from_offset + offset_shift
                to_offset = from_offset_shift + len(target)
                offsets = [(from_offset_shift, to_offset)]
                offset_text = tweet_text[from_offset_shift : to_offset].lower()
                if offset_text == target.lower():
                    return offsets
            raise ValueError('Offset {} does not match target text {}. Full '\
                             'text {}\nid {}'\
                             .format(from_offset, target, tweet_text, tweet_id))

        def fuzzy_target_match(tweet_text, target):
            low_target = target.lower()
            target_searches = [low_target, r'[^\w]' + low_target,
                               r'[^\w]' + low_target + r'[^\w]',
                               low_target + r'[^\w]',
                               low_target.replace(' ', ''),
                               low_target.replace(" '", '')]
            for target_search in target_searches:
                target_matches = list(re.finditer(target_search,
                                                  tweet_text.lower()))
                if len(target_matches) == 1:
                    return target_matches
            if tweet_id in set(['81211671026352128', '78689580104290305',
                                '81209490499960832']):
                return None
            if tweet_id == '75270720671973376' and target == 'kippers':
                return None
            if tweet_id == '65855178264686592' and target == 'tax':
                return None
            print(tweet_data)
            print(anno_data)
            raise ValueError('Cannot find the exact additional '\
                             'entity {} within the tweet {}'\
                             .format(target, tweet_text))



        target_instances = []
        tweet_id = str(tweet_id)
        tweet_text = tweet_data['content']
        target_ids = []
        # Parse all of the entities that have been detected automatically
        for entity in tweet_data['entities']:
            data_dict = {}
            target = entity['entity']
            target_ids.append(entity['id'])
            entity_id = str(entity['id'])
            data_dict['spans'] = get_offsets(entity, tweet_text, target)
            data_dict['target'] = entity['entity']
            data_dict['target_id'] = folder_name + tweet_id + '#' + entity_id
            data_dict['sentence_id'] = folder_name + tweet_id
            data_dict['sentiment'] = anno_data['items'][entity_id]
            if data_dict['sentiment'] == 'doesnotapply' and not include_dnr:
                continue
            # Convert from Strings to Integer
            data_dict['sentiment'] = sentiment_mapper[data_dict['sentiment']]
            data_dict['text'] = tweet_text
            target_instances.append(Target(**data_dict))
        # Parse all of the entities that have been selected by the user
        if include_additional:
            additional_data = anno_data['additional_items']
            if isinstance(additional_data, dict):
                for target, sentiment in additional_data.items():
                    target_matches = fuzzy_target_match(tweet_text, target)
                    if target_matches is None:
                        continue
                    target_id = max(target_ids) + 1
                    target_ids.append(target_id)
                    data_dict['spans'] = [target_matches[0].span()]
                    data_dict['target'] = target
                    data_dict['sentiment'] = sentiment
                    data_dict['text'] = tweet_text
                    data_dict['sentence_id'] = tweet_id
                    data_dict['target_id'] = tweet_id + '#' + str(target_id)
                    target_instances.append(Target(**data_dict))

        return target_instances

    def get_data(id_file, tweets_data, annos_data):
        targets = []
        with open(id_file, 'r') as id_data:
            for tweet_id in id_data:
                tweet_id = tweet_id.strip()
                tweet_data = tweets_data[tweet_id]
                anno_data = annos_data[tweet_id]
                targets.extend(parse_tweet(tweet_data, anno_data, tweet_id))
        return TargetCollection(targets)

    tweets_data = get_file_data(os.path.join(folder_path, 'tweets'))
    annotations_data = get_file_data(os.path.join(folder_path, 'annotations'))

    train_ids_file = os.path.join(folder_path, 'train_id.txt')
    train_data = get_data(train_ids_file, tweets_data, annotations_data)
    test_ids_file = os.path.join(folder_path, 'test_id.txt')
    test_data = get_data(test_ids_file, tweets_data, annotations_data)

    return train_data, test_data

def election_train(folder_path: Path, include_dnr: bool = False, 
                   include_additional: bool = False, **target_collection_kwargs
                   ) -> TargetCollection:
    '''
    Data can be downloaded from
    `FigShare <https://figshare.com/articles/EACL_2017_-_Multi-target_\
    UK_election_Twitter_sentiment_corpus/4479563/1>`_. This function parses 
    and returns only the training data.

    :param folder_path: Path to the folder containing the data after it has
                        been unziped and all folders within it have been 
                        unziped.
    :param include_dnr: determine if to include the `doesnotapply` label
    :param include_additional: NOTE: This does not work at the moment. 
                               Determine if to parse the additional data.
    :param target_collection_kwargs: Keywords to parse to the TargetCollection 
                                     constructor that is returned.
    :returns: A TargetCollection containing Target instances.
    '''

    train_data, _ = election(folder_path, include_dnr=include_dnr, 
                             include_additional=include_additional)
    return TargetCollection(train_data.data(), **target_collection_kwargs)

def election_test(folder_path: Path, include_dnr: bool = False, 
                   include_additional: bool = False, **target_collection_kwargs
                   ) -> TargetCollection:
    '''
    Data can be downloaded from
    `FigShare <https://figshare.com/articles/EACL_2017_-_Multi-target_\
    UK_election_Twitter_sentiment_corpus/4479563/1>`_. This function parses 
    and returns only the test data.

    :param folder_path: Path to the folder containing the data after it has
                        been unziped and all folders within it have been 
                        unziped.
    :param include_dnr: determine if to include the `doesnotapply` label
    :param include_additional: NOTE: This does not work at the moment. 
                               Determine if to parse the additional data.
    :param target_collection_kwargs: Keywords to parse to the TargetCollection 
                                     constructor that is returned.
    :returns: A TargetCollection containing Target instances.
    '''

    _, test_data = election(folder_path, include_dnr=include_dnr, 
                            include_additional=include_additional)
    return TargetCollection(test_data.data(), **target_collection_kwargs)

def hu_liu(file_path):
    '''
    Parser for the datasets from the following two papers (DOES NOT WORK):

    1. `A Holistic Lexicon-Based Approach to Opinion Mining \
    <https://www.cs.uic.edu/~liub/FBS/opinion-mining-final-WSDM.pdf>`_
    2. `Mining and Summarizing Customer Reviews \
    <https://www.cs.uic.edu/~liub/publications/kdd04-revSummary.pdf>`_

    Currently this does not work. This is due to the dataset not containing
    enough data to determine where the targets are in the text.

    :param file_path: The path to a file containing annotations in the format \
    of hu and liu sentiment datasets.
    :type file_path: String
    :returns: A TargetCollection containing Target instances.
    :rtype: TargetCollection
    '''
    file_path = os.path.abspath(file_path)
    file_name = os.path.basename(file_path)
    sentiment_data = TargetCollection()

    with open(file_path, 'r', encoding='cp1252') as annotations:
        for sentence_index, annotation in enumerate(annotations):
            # If it does not contain ## then not a sentence
            if '##' not in annotation:
                continue
            targets_text = annotation.split('##')
            if len(targets_text) > 2 or len(targets_text) < 1:
                raise ValueError('The annotation {} when split on `##` should '\
                                 'contain at least the sentence text and at'\
                                 ' most the text and the targets and not {}'\
                                 .format(annotation, targets_text))
            # If it just contains the sentence text then go to next
            elif len(targets_text) == 1:
                continue
            elif targets_text[0].strip() == '':
                continue
            targets, text = targets_text
            targets = targets.strip()
            text = text.strip()
            sentence_id = file_name + '#{}'.format(sentence_index)

            targets = targets.split(',')
            for target_index, target in enumerate(targets):
                target = target.strip()
                sentiment_match = re.search(r'\[[+-]\d\]$', target)
                is_implicit = re.search(r'\[[up]\]', target)
                if is_implicit:
                    print('Target {} is implicit {}'.format(target, text))
                    continue
                if not sentiment_match:
                    raise ValueError('Target {} does not have a corresponding'\
                                     ' sentiment value. annotation {}'\
                                     .format(target, annotation))
                target_text = target[:sentiment_match.start()].strip()
                sentiment_text = sentiment_match.group().strip().strip('[]')
                sentiment_value = int(sentiment_text)

                target_matches = list(re.finditer(target_text, text))
                if len(target_matches) != 1:
                    print('The Target {} can only occur once in the '\
                          'text {}'.format(target_text, text))
                    continue
                    raise ValueError('The Target {} can only occur once in the '\
                                     'text {}'.format(target_text, text))
                target_span = target_matches[0].span()
                target_id = sentence_id + '#{}'.format(target_index)

                data_dict = {}
                data_dict['spans'] = [target_span]
                data_dict['target'] = target_text
                data_dict['sentiment'] = sentiment_value
                data_dict['text'] = text
                data_dict['sentence_id'] = sentence_id
                data_dict['target_id'] = target_id
                sentiment_data.add(Target(**data_dict))
    return sentiment_data


def mitchel(file_name: Path, **target_collection_kwargs) -> TargetCollection:
    '''
    Parser for the dataset introduced by `Mitchel et al. \
    <https://www.aclweb.org/anthology/D13-1171>`_. The dataset can be downloaded
    from `<here http://www.m-mitchell.com/code/MitchellEtAl-13-OpenSentiment.tgz>`_
    the dataset can be found within the tarball under /en/10-fold and then
    choose one of the folds e.g. train_1 and test_1 to get the full dataset.

    :param file_path: path to either the train or test data.
    :param target_collection_kwargs: Keywords to parse to the TargetCollection 
                                     constructor that is returned.
    :returns: A TargetCollection containing Target instances.
    '''

    def extract_targets(current_target, end_span, start_span, targets,
                        target_spans, target_index, tweet_text, sentiment_data,
                        tweet_id, target_sentiments):
        if current_target != []:
            target_word = ' '.join(current_target)
            end_span = start_span + len(target_word)
            targets.append(target_word)
            target_spans.append((start_span, end_span))
            start_span, end_span = None, None
            current_target = []
            target_index += 1
        tweet_text = ' '.join(tweet_text)
        for index, target in enumerate(targets):
            target_id = '{}#{}'.format(tweet_id, index)
            target_sentiment = target_sentiments[index]
            target_span = target_spans[index]
            if tweet_text[target_span[0] : target_span[1]] != target:
                raise Exception('The target span {} does not match the '\
                                'target word {} in {}'\
                                .format(target_span, target, tweet_text))
            target_data = {'spans' : [target_span], 'target_id' : target_id,
                           'target' : target, 'text' : tweet_text,
                           'sentiment' : target_sentiment,
                           'sentence_id' : tweet_id}
            target_data = Target(**target_data)
            sentiment_data.add(target_data)

        return sentiment_data

    sentiment_mapper = {'negative' : -1, 'neutral' : 0, 'positive' : 1}
    sentiment_data = TargetCollection(**target_collection_kwargs)

    with open(file_name, 'r') as fp:
        tweet_id = None
        tweet_text = []
        targets = []
        current_target = []
        target_sentiments = []
        target_spans = []
        start_span = None
        end_span = None
        target_index = 0

        for line in fp:
            line = line.strip()
            tweet_id_line = re.match(r'## Tweet (\d+)', line)
            if tweet_id_line is not None:

                if tweet_text != []:
                    sentiment_data = extract_targets(current_target, end_span,
                                                     start_span, targets,
                                                     target_spans,
                                                     target_index, tweet_text,
                                                     sentiment_data, tweet_id,
                                                     target_sentiments)
                    tweet_text = []
                    targets = []
                    current_target = []
                    target_sentiments = []
                    target_spans = []
                    target_index = 0
                    start_span = None
                    end_span = None
                    target_index = 0
                tweet_id = tweet_id_line.group(1)
                continue

            if line == '':
                continue

            line_data = line.split('\t')
            if len(line_data) != 3:
                if len(line_data) == 4:
                    if line_data[2] == 'NUMBER':
                        line_data = line_data[0], line_data[1], line_data[3]
                    else:
                        raise Exception('Cannot parse line {} in Tweet ID {}'\
                                        .format(line, tweet_id))
                else:
                    raise Exception('Cannot parse line {} in Tweet ID {}'.format(line, tweet_id))
            word, ner_data, sentiment = line_data
            if len(word.split()) != 1:
                raise Exception('Why is the word got a space in it: {}'.format(word))
            word = ftfy.fix_encoding(word)
            tweet_text.append(word)

            # Contains sentiment
            if sentiment != '_':
                if len(current_target) != 0:
                    if ner_data[0] == 'B':
                        target_word = ' '.join(current_target)
                        end_span = start_span + len(target_word)
                        targets.append(target_word)
                        target_spans.append((start_span, end_span))
                        start_span, end_span = None, None
                        current_target = []
                        target_index += 1
                    else:
                        raise Exception('Contains the following target {} id {}'\
                                        .format(current_target, tweet_id))
                current_target.append(word)
                sentiment = sentiment_mapper[sentiment]
                target_sentiments.append(sentiment)
                start_span = len(' '.join(tweet_text)) - len(word)
            elif len(current_target) != 0:
                if ner_data[0] == 'I':
                    current_target.append(word)
                else:
                    target_word = ' '.join(current_target)
                    end_span = start_span + len(target_word)
                    targets.append(target_word)
                    target_spans.append((start_span, end_span))
                    start_span, end_span = None, None
                    current_target = []
                    target_index += 1
        sentiment_data = extract_targets(current_target, end_span, start_span,
                                         targets, target_spans, target_index,
                                         tweet_text, sentiment_data, tweet_id,
                                         target_sentiments)
    return sentiment_data
